# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/04_plots.ipynb.

# %% auto 0
__all__ = ['plot_losses', 'plot_comparision', 'new_plot_comparisions', 'plot_gene_trends']

# %% ../nbs/04_plots.ipynb 3
import os, math, numpy as np, pandas as pd
import torch
import torch.nn as nn

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
sns.color_palette("bright")
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D
from .utils import to_np
from .constants import IMGS_DIR
def plot_losses(local_losses=None, batch_losses=None, globe_losses=None, save=False, path=IMGS_DIR, file='losses.png'):
    if not os.path.isdir(path):
        os.makedirs(path)

    s = 1
    fig = plt.figure(figsize=(14/s, 6/s), dpi=300)
    cur = 0
    sub = sum([1 for e in [local_losses, batch_losses, globe_losses] if e is not None])
    
    if local_losses is not None:
        cur += 1
        ax = fig.add_subplot(1,sub,cur)

        df_lloss = pd.DataFrame(local_losses)

        ax.set_title('Loss per step')
        sns.lineplot(data=df_lloss, ax=ax)
        
    
    if batch_losses is not None:
        cur += 1        
        ax = fig.add_subplot(1,sub,cur)
        
        df_bloss = pd.DataFrame(batch_losses, columns=['Batch Loss'])        
        
        ax.set_title('Batch loss per batch')        
        sns.lineplot(data=df_bloss, lw=1, ax=ax)
        
    if globe_losses is not None:
        cur += 1
        ax = fig.add_subplot(1,sub,cur)
        
        df_gloss = pd.DataFrame(globe_losses, columns=['Global Loss'])        

        ax.set_title('Global Loss per batch');
        sns.lineplot(data=df_gloss, ax=ax)
    
    if save:
        # NOTE: savefig complains image is too large but saves it anyway. 
        try:
            fig.savefig(os.path.expanduser(os.path.join(path, file)))
        except ValueError:
            pass
    plt.close()
    return fig

# %% ../nbs/04_plots.ipynb 4
import os, math, numpy as np, pandas as pd
import torch
import torch.nn as nn

import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
sns.color_palette("bright")
import matplotlib as mpl
import matplotlib.cm as cm
from mpl_toolkits.mplot3d import Axes3D
from .utils import to_np
from .constants import IMGS_DIR
def plot_comparision(
    df, generated, trajectories,
    palette = 'viridis', df_time_key='samples',
    save=False, path=IMGS_DIR, file='comparision.png',
    x='d1', y='d2', z='d3', is_3d=False
):
    if not os.path.isdir(path):
        os.makedirs(path)

    if not is_3d:
        return new_plot_comparisions(
            df, generated, trajectories,
            palette=palette, df_time_key=df_time_key,
            x=x, y=y, z=z, is_3d=is_3d,            
            groups=None,
            save=save, path=path, file=file,
        )

    s = 1
    fig = plt.figure(figsize=(12/s, 8/s), dpi=300)
    if is_3d:
        ax = fig.add_subplot(1,1,1,projection='3d')
    else:
        ax = fig.add_subplot(1,1,1)
    
    states = sorted(df[df_time_key].unique())
    
    if is_3d:
        ax.scatter(
            df[x], df[y], df[z],
            cmap=palette, alpha=0.3,
            c=df[df_time_key], 
            s=df[df_time_key], 
            marker='X',
        )
    else:
        sns.scatterplot(
            data=df, x=x, y=y, palette=palette, alpha=0.3,
            hue=df_time_key, style=df_time_key, size=df_time_key,
            markers={g: 'X' for g in states},
            sizes={g: 100 for g in states}, 
            ax=ax, legend=False
        )
    

    if not isinstance(generated, np.ndarray):
        generated = to_np(generated)
    points = np.concatenate(generated, axis=0)
    n_gen = int(points.shape[0] / len(states))
    colors = [state for state in states for i in range(n_gen)]
    
    if is_3d:
        ax.scatter(
            points[:, 0], points[:, 1], points[:, 2],
            cmap=palette,
            c=colors, 
        )
    else:
        sns.scatterplot(
            x=points[:, 0], y=points[:, 1], palette=palette,
            hue=colors, 
            ax=ax, legend=False
        )
    
    ax.legend(title='Timepoint', loc='upper left', labels=['Ground Truth', 'Predicted'])
    ax.set_title('ODE Points compared to Ground Truth')

    if is_3d:
        for trajectory in np.transpose(trajectories, axes=(1,0,2)):
            plt.plot(trajectory[:, 0], trajectory[:, 1], trajectory[:, 2], alpha=0.1, color='Black');
    else:
        for trajectory in np.transpose(trajectories, axes=(1,0,2)):
            plt.plot(trajectory[:, 0], trajectory[:, 1], alpha=0.1, color='Black');
        
    if save:
        # NOTE: savefig complains image is too large but saves it anyway. 
        try:
            fig.savefig(os.path.expanduser(os.path.join(path, file)))
        except ValueError:
            pass 
    plt.close()
    return fig

# %% ../nbs/04_plots.ipynb 5
import matplotlib.pyplot as plt
from matplotlib.patches import Patch
from matplotlib.lines import Line2D
from matplotlib import rcParams, cycler

def new_plot_comparisions(
    df, generated, trajectories,
    palette = 'viridis',
    df_time_key='samples',
    x='d1', y='d2', z='d3', 
    groups=None,
    save=False, path=IMGS_DIR, file='comparision.png',
    is_3d=False
):
    if groups is None:
        groups = sorted(df[df_time_key].unique())
    cmap = plt.cm.viridis
    sns.set_palette(palette)
    plt.rcParams.update({
        'axes.prop_cycle': plt.cycler(color=cmap(np.linspace(0, 1, len(groups) + 1))),
        'axes.axisbelow': False,
        'axes.edgecolor': 'lightgrey',
        'axes.facecolor': 'None',
        'axes.grid': False,
        'axes.labelcolor': 'dimgrey',
        'axes.spines.right': False,
        'axes.spines.top': False,
        'figure.facecolor': 'white',
        'lines.solid_capstyle': 'round',
        'patch.edgecolor': 'w',
        'patch.force_edgecolor': True,
        'text.color': 'dimgrey',
        'xtick.bottom': False,
        'xtick.color': 'dimgrey',
        'xtick.direction': 'out',
        'xtick.top': False,
        'ytick.color': 'dimgrey',
        'ytick.direction': 'out',
        'ytick.left': False,
        'ytick.right': False, 
        'font.size':12, 
        'axes.titlesize':10,
        'axes.labelsize':12
    })

    n_cols = 1
    n_rols = 1

    grid_figsize = [12, 8]
    dpi = 300
    grid_figsize = (grid_figsize[0] * n_cols, grid_figsize[1] * n_rols)
    fig = plt.figure(None, grid_figsize, dpi=dpi)

    hspace = 0.3
    wspace = None
    gspec = plt.GridSpec(n_rols, n_cols, fig, hspace=hspace, wspace=wspace)

    outline_width = (0.3, 0.05)
    size = 300
    bg_width, gap_width = outline_width
    point = np.sqrt(size)

    gap_size = (point + (point * gap_width) * 2) ** 2
    bg_size = (np.sqrt(gap_size) + (point * bg_width) * 2) ** 2

    plt.legend(frameon=False)

    is_3d = False
    
    # if is_3d:        
    #     ax = fig.add_subplot(1,1,1,projection='3d')
    # else:
    #     ax = fig.add_subplot(1,1,1)
    
    axs = []
    for i, gs in enumerate(gspec):        
        ax = plt.subplot(gs)
        
        
        n = 0.3   
        ax.scatter(
                df[x], df[y],
                c=df[df_time_key],
                s=size,
                alpha=0.7 * n,
                marker='X',
                linewidths=0,
                edgecolors=None,
                cmap=cmap
            )
        
        for trajectory in np.transpose(trajectories, axes=(1,0,2)):
                plt.plot(trajectory[:, 0], trajectory[:, 1], alpha=0.3, color='Black');
        
        states = sorted(df[df_time_key].unique())
        points = np.concatenate(generated, axis=0)
        n_gen = int(points.shape[0] / len(states))
        colors = [state for state in states for i in range(n_gen)]
        n = 1
        o = '.'
        ax.scatter(
                points[:, 0], points[:, 1],
                c='black',
                s=bg_size,
                alpha=1 * n,
                marker=o,
                linewidths=0,
                edgecolors=None
            )
        ax.scatter(
                points[:, 0], points[:, 1],
                c='white',
                s=gap_size,
                alpha=1 * n,
                marker=o,
                linewidths=0,
                edgecolors=None
            )
        pnts = ax.scatter(
                points[:, 0], points[:, 1],
                c=colors,
                s=size,
                alpha=0.7 * n,
                marker=o,
                linewidths=0,
                edgecolors=None,
                cmap=cmap
            )
                
        legend_elements = [        
            Line2D(
                [0], [0], marker='o', 
                color=cmap((i) / (len(states)-1)), label=f'T{state}', 
                markerfacecolor=cmap((i) / (len(states)-1)), markersize=15,
            )
            for i, state in enumerate(states)
        ]
        
        leg = plt.legend(handles=legend_elements, loc='upper left')
        ax.add_artist(leg)
        
        legend_elements = [        
            Line2D(
                [0], [0], marker='X', color='w', 
                label='Ground Truth', markerfacecolor=cmap(0), markersize=15, alpha=0.3
            ),
            Line2D([0], [0], marker='o', color='w', label='Predicted', markerfacecolor=cmap(.999), markersize=15),
            Line2D([0], [0], color='black', lw=2, label='Trajectory')
            
        ]
        leg = plt.legend(handles=legend_elements, loc='upper right')
        ax.add_artist(leg)
        
        ax.set_xlabel("")
        ax.set_ylabel("")
        ax.get_xaxis().get_major_formatter().set_scientific(False)
        ax.get_yaxis().get_major_formatter().set_scientific(False)
        kwargs = dict(bottom=False, left=False, labelbottom=False, labelleft=False)
        ax.tick_params(which="both", **kwargs)
        ax.set_frame_on(False)
        ax.patch.set_alpha(0)
        

        axs.append(ax)

    if save:
        # NOTE: savefig complains image is too large but saves it anyway. 
        try:
            fig.savefig(os.path.expanduser(os.path.join(path, file)))
        except ValueError:
            pass 
    plt.close()
    return fig

# %% ../nbs/04_plots.ipynb 6
import matplotlib.patches as mpatches
import numpy as np, matplotlib.pyplot as plt
from .utils import get_times_from_groups
def plot_gene_trends(
    genes, top_idxs, inverse, colors, 
    samples=None, groups=None,
    n_cols=None, n_rows=None, where='end', start=0, top_n=None,
    cell_types=None, use_cell_types=True, 
    save:bool=False, path:str=IMGS_DIR, file:str='gene_trends.png'
):
    '''
    Notes:
        - first four arguments `genes`, `top_idxs`, `inverse`, and `colors` are output from `get_cell_indexes`
            see that function's docstring for more information.
    
    Arguments:
    ----------
        genes (np.ndarray): List of genes similar to those the user passed into this function except
            in order of the columns of the provided `df`. Any genes not found in the `df` put passed in
            by the user will be removed.
            
        top_idxs (dict | dict[dict]): See notes. Dictionary or nested dictionary where leaf values are
            indicies of cells corresponding to those expressing the highest amount of specified genes.
            
        inverse (np.nddary): Reconstructed gene space from `trajectories` and `principal_components`.
            It has the shape (n_time * n_bins, n_cells, n_genes). See `generate_tjnet_trajectories`.
        
        colors (dict): Dictionary of either `{gene: color}` or `{cell_type: color}` depending on `use_cell_types`.
    
        samples (np.ndarray | list): List of timepoints where each value corresponds to the 
            timepoint of the same row in `df`. Defaults to `None`.
        
        groups (np.ndarray | list): List of time groups in order (e.g. `[0, 1, 2, 3, 4, 5, 6, 7]`).
            Defaults to `None`.
            
        n_cols (int): How many columns to use. Defaults to `None`. If `n_cols == n_rows == None`
            then `n_cols = 4` and `n_rows` is calculated dynamically.
        
        n_rows (int): How many rows to use. Defaults to `None`. If `n_cols == n_rows == None`
            then `n_cols = 4` and `n_rows` is calculated dynamically.
        
        where (str): Choices are `"start"`, and `"end"`. Defaults to `"end"`. Whether or not
            the trajectories used to calculate `top_idxs` start at 
            `t_0` (`"start"`) or `t_n` (`"end"`). 

        start (int): Defaults to `0`. Where in `generate_tjnet_trajectories` the trajectories started.
            This is used if attempting to generate outside of `t0`. Note this works relative to `where`.
            E.g. if `where="end"` and `start=0` then this is the same as `groups[-1]`.
            
        top_n (int): Defaults to `None`. Used for legend. If not provided will be wrangled from `top_idxs`.
        
        cell_types (np.ndarray | list): List of cell types to plot. Should be outer keys from `top_idxs`.
            This can be a subset of those keys to plot less.
        
        use_cell_types (bool): Whether or not to use cell types.

        save (bool): Whether or not to save the file. Defaults to `False`.

        path (str): Path to the directory in which to save the file. Defaults to `IMGS_DIR`.

        file (str): Name of the file to be saved in directory specified by `path`. Defaults
            to `"gene_trends.png"`
        
    Returns:
    ----------
        fig (matplotlib.pyplot.figure)
    '''
    
    # Test for valid start location    
    _valid_where = 'start end'.split()
    if where not in _valid_where:
        raise ValueError(f'{where} not known. Should be one of {_valid_where}')
        
    if groups is None and samples is None:
        raise ValueError(f'`groups == samples == None`. `groups` or `samples` must not be `None`.')
        
    # Try to figure out top_n if not provided
    if top_n is None:
        top_n = int(np.round(np.mean(list(map(len, top_idxs.values())))))
    
    # Get groups from samples
    if groups is None:
        groups = sorted(np.unique(samples))

    # NOTE: update to handle generation other than t0
    # times = groups
    # if where == 'end':
    #     times = times[::-1]
    # times = times[start:]
    # groups = times
    groups = get_times_from_groups(groups, where, start)

    # Figure out columns / rows from number of genes
    n_genes = len(genes)
    if n_cols is None and n_rows is None:
        n_cols = 4
        n_rows = np.ceil(n_genes / n_cols).astype(int)
    elif n_cols is None:
        n_cols = np.ceil(n_genes / n_rows).astype(int)
    elif n_rows is None:
        n_rows = np.ceil(n_genes / n_cols).astype(int)
    else:
        pass
    
    # BEGIN: plot options
    grid_figsize = [4, 3]
    
    dpi = 300
    grid_figsize = (grid_figsize[0] * n_cols, grid_figsize[1] * n_rows)
    fig = plt.figure(None, grid_figsize, dpi=dpi)
    
    hspace = 0.3
    wspace = None
    gspec = plt.GridSpec(n_rows, n_cols, fig, hspace=hspace, wspace=wspace)
        
    plt.legend(frameon=False)
    # END: plot options
    
    
    axs = []
    for i, gs in enumerate(gspec):
        # Grid may have more subplots than genes to plot
        if i >= len(genes):
            break
                 
        ax = plt.subplot(gs)
        
        # current gene subplot
        gene = genes[i]
                
        ax.set_title(gene)
        ax.set_yticks([])
        ax.set_xticks(range(np.min(groups), np.max(groups)))
        
        # x-axis 0 ---> n or n ---> 0 depending on `where`
        x_axis = np.linspace(np.min(groups), np.max(groups), inverse.shape[0])
        x_axis = x_axis if where == 'start' else x_axis[::-1]
                    
        if use_cell_types:
            for j, (ctype, gexpr) in enumerate(top_idxs.items()):
                if ctype not in cell_types:
                    continue                
                for k, cell_idx in enumerate(gexpr[gene]):    
                    # Plot top cells of given cell type expressing given gene
                    ax.plot(x_axis, inverse[:, cell_idx, i], c=colors[ctype], label=f'{ctype}_{gene}')    
                    
        else:
            for j, eg in enumerate(genes):
                for cell in top_idxs[eg]:     
                    # Plot top cells expressing any specified genes, expressing given gene
                    ax.plot(x_axis, inverse[:, cell, i], c=colors[eg], label=gene)

    
    if use_cell_types:
        fig.legend(
            handles=[
                mpatches.Patch(color=colors[ctype], label=ctype)
                for ctype in cell_types
            ], 
            title=f'Cell Types'
        )        
    else:
        fig.legend(
            handles=[
                mpatches.Patch(color=colors[gene], label=gene)
                for gene in genes
            ], 
            title=f'Top {top_n} Cells Expressing Genes', 
            bbox_to_anchor=(0, 0, 0, 0.5)
        )

    if save:
        # NOTE: savefig complains image is too large but saves it anyway. 
        try:
            fig.savefig(os.path.expanduser(os.path.join(path, file)))
        except ValueError:
            pass         
        
    plt.close()
    return fig



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/00_ode.ipynb.

# %% auto 0
__all__ = ['ode_solve', 'ODEF', 'ODEAdjoint', 'NeuralODE']

# %% ../nbs/00_ode.ipynb 3
import os, math, numpy as np
import torch
import torch.nn as nn

def ode_solve(z0, t0, t1, f):
    """
    Simplest Euler ODE initial value solver
    """
    h_max = 0.05
    n_steps = math.ceil((abs(t1 - t0)/h_max).max().item())

    h = (t1 - t0)/n_steps
    t = t0
    z = z0

    for i_step in range(n_steps):
        z = z + h * f(z, t)
        t = t + h
    return z

class ODEF(nn.Module):
    def forward_with_grad(self, z, t, grad_outputs):
        """Compute f and a df/dz, a df/dp, a df/dt"""
        batch_size = z.shape[0]

        out = self.forward(z, t)

        a = grad_outputs
        adfdz, adfdt, *adfdp = torch.autograd.grad(
            (out,), (z, t) + tuple(self.parameters()), grad_outputs=(a),
            allow_unused=True, retain_graph=True
        )
        # grad method automatically sums gradients for batch items, we have to expand them back 
        if adfdp is not None:
            adfdp = torch.cat([p_grad.flatten() for p_grad in adfdp]).unsqueeze(0)
            adfdp = adfdp.expand(batch_size, -1) / batch_size
        if adfdt is not None:
            adfdt = adfdt.expand(batch_size, 1) / batch_size
        return out, adfdz, adfdt, adfdp

    def flatten_parameters(self):
        p_shapes = []
        flat_parameters = []
        for p in self.parameters():
            p_shapes.append(p.size())
            flat_parameters.append(p.flatten())
        return torch.cat(flat_parameters)
    
class ODEAdjoint(torch.autograd.Function):
    @staticmethod
    def forward(ctx, z0, t, flat_parameters, func):
        assert isinstance(func, ODEF)
        bs, *z_shape = z0.size()
        time_len = t.size(0)

        with torch.no_grad():
            z = torch.zeros(time_len, bs, *z_shape).to(z0)
            z[0] = z0
            for i_t in range(time_len - 1):
                z0 = ode_solve(z0, t[i_t], t[i_t+1], func)
                z[i_t+1] = z0

        ctx.func = func
        ctx.save_for_backward(t, z.clone(), flat_parameters)
        return z

    @staticmethod
    def backward(ctx, dLdz):
        """
        dLdz shape: time_len, batch_size, *z_shape
        """
        func = ctx.func
        t, z, flat_parameters = ctx.saved_tensors
        time_len, bs, *z_shape = z.size()
        n_dim = np.prod(z_shape)
        n_params = flat_parameters.size(0)

        # Dynamics of augmented system to be calculated backwards in time
        def augmented_dynamics(aug_z_i, t_i):
            """
            tensors here are temporal slices
            t_i - is tensor with size: bs, 1
            aug_z_i - is tensor with size: bs, n_dim*2 + n_params + 1
            """
            z_i, a = aug_z_i[:, :n_dim], aug_z_i[:, n_dim:2*n_dim]  # ignore parameters and time

            # Unflatten z and a
            z_i = z_i.view(bs, *z_shape)
            a = a.view(bs, *z_shape)
            with torch.set_grad_enabled(True):
                t_i = t_i.detach().requires_grad_(True)
                z_i = z_i.detach().requires_grad_(True)
                func_eval, adfdz, adfdt, adfdp = func.forward_with_grad(z_i, t_i, grad_outputs=a)  # bs, *z_shape
                adfdz = adfdz.to(z_i) if adfdz is not None else torch.zeros(bs, *z_shape).to(z_i)
                adfdp = adfdp.to(z_i) if adfdp is not None else torch.zeros(bs, n_params).to(z_i)
                adfdt = adfdt.to(z_i) if adfdt is not None else torch.zeros(bs, 1).to(z_i)

            # Flatten f and adfdz
            func_eval = func_eval.view(bs, n_dim)
            adfdz = adfdz.view(bs, n_dim) 
            return torch.cat((func_eval, -adfdz, -adfdp, -adfdt), dim=1)

        dLdz = dLdz.view(time_len, bs, n_dim)  # flatten dLdz for convenience
        with torch.no_grad():
            ## Create placeholders for output gradients
            # Prev computed backwards adjoints to be adjusted by direct gradients
            adj_z = torch.zeros(bs, n_dim).to(dLdz)
            adj_p = torch.zeros(bs, n_params).to(dLdz)
            # In contrast to z and p we need to return gradients for all times
            adj_t = torch.zeros(time_len, bs, 1).to(dLdz)

            for i_t in range(time_len-1, 0, -1):
                z_i = z[i_t]
                t_i = t[i_t]
                f_i = func(z_i, t_i).view(bs, n_dim)

                # Compute direct gradients
                dLdz_i = dLdz[i_t]
                dLdt_i = torch.bmm(torch.transpose(dLdz_i.unsqueeze(-1), 1, 2), f_i.unsqueeze(-1))[:, 0]

                # Adjusting adjoints with direct gradients
                adj_z += dLdz_i
                adj_t[i_t] = adj_t[i_t] - dLdt_i

                # Pack augmented variable
                aug_z = torch.cat((z_i.view(bs, n_dim), adj_z, torch.zeros(bs, n_params).to(z), adj_t[i_t]), dim=-1)

                # Solve augmented system backwards
                aug_ans = ode_solve(aug_z, t_i, t[i_t-1], augmented_dynamics)

                # Unpack solved backwards augmented system
                adj_z[:] = aug_ans[:, n_dim:2*n_dim]
                adj_p[:] += aug_ans[:, 2*n_dim:2*n_dim + n_params]
                adj_t[i_t-1] = aug_ans[:, 2*n_dim + n_params:]

                del aug_z, aug_ans

            ## Adjust 0 time adjoint with direct gradients
            # Compute direct gradients 
            dLdz_0 = dLdz[0]
            dLdt_0 = torch.bmm(torch.transpose(dLdz_0.unsqueeze(-1), 1, 2), f_i.unsqueeze(-1))[:, 0]

            # Adjust adjoints
            adj_z += dLdz_0
            adj_t[0] = adj_t[0] - dLdt_0
        return adj_z.view(bs, *z_shape), adj_t, adj_p, None
    
class NeuralODE(nn.Module):
    def __init__(self, func):
        super(NeuralODE, self).__init__()
        assert isinstance(func, ODEF)
        self.func = func

    def forward(self, z0, t=torch.Tensor([0., 1.]), return_whole_sequence=False):
        t = t.to(z0)
        z = ODEAdjoint.apply(z0, t, self.func.flatten_parameters(), self.func)
        if return_whole_sequence:
            return z
        else:
            return z[-1]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_constants.ipynb.

# %% auto 0
__all__ = ['ROOT_DIR', 'DATA_DIR', 'NTBK_DIR', 'IMGS_DIR', 'RES_DIR', 'WORM_FILE', 'EB_BODIES_FILE', 'EB_BODIES_PSEUDO_4',
           'EB_BODIES_PSEUDO_6', 'EB_BODIES_PSEUDO_25', 'EB_BODIES_PSEUDO_82', 'DYNGEN_INFO_FILE', 'DYNGEN_EXPR_FILE']

# %% ../nbs/06_constants.ipynb 3
import os, inspect
__file = inspect.getfile(lambda: None)
ROOT_DIR = os.path.abspath(os.path.join(os.path.dirname(__file), '..'))
DATA_DIR = os.path.join(ROOT_DIR, 'data')
NTBK_DIR = os.path.join(ROOT_DIR, 'notebooks')
IMGS_DIR = os.path.join(ROOT_DIR, 'imgs')
RES_DIR = os.path.join(ROOT_DIR, 'results')
WORM_FILE = os.path.join(DATA_DIR, 'worm_TrNet2.npz')
EB_BODIES_FILE = os.path.join(DATA_DIR, 'natalia_eb_rna_smoothed.npz')
EB_BODIES_PSEUDO_4 = os.path.join(DATA_DIR, 'pseudotime-4x.npy')
EB_BODIES_PSEUDO_6 = os.path.join(DATA_DIR, 'pseudotime-6x.npy')
EB_BODIES_PSEUDO_25 = os.path.join(DATA_DIR, 'pseudotime-25x.npy')
EB_BODIES_PSEUDO_82 = os.path.join(DATA_DIR, 'pseudotime-82x.npy')

DYNGEN_INFO_FILE = os.path.join(DATA_DIR, 'cell_info.csv')
DYNGEN_EXPR_FILE = os.path.join(DATA_DIR, 'dyngen_expression_bif.csv')



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/07_datasets.ipynb.

# %% auto 0
__all__ = ['construct_diamond', 'make_diamonds', 'make_swiss_roll', 'make_tree', 'make_worm_data', 'make_eb_data',
           'make_dyngen_data', 'relabel_data', 'rings', 'make_rings', 'make_jacks', 'branch_data_clean',
           'branch_data_data', 'make_branch_cond', 'make_partial_branch', 'make_uniform_rect', 'make_gaussian_rect',
           'make_dying_example_unif', 'make_dying_example_gaus', 'make_growing']

# %% ../nbs/07_datasets.ipynb 3
import os
import pandas as pd, numpy as np
import phate
from sklearn import datasets

import seaborn as sns
sns.color_palette("bright")
import matplotlib as mpl

# %% ../nbs/07_datasets.ipynb 4
def construct_diamond(
    points_per_petal:int=200,
    petal_width:float=0.25,
    direction:str='y'
):
    '''
    Arguments:
    ----------
        points_per_petal (int). Defaults to `200`. Number of points per petal.
        petal_width (float): Defaults to `0.25`. How narrow the diamonds are.
        direction (str): Defaults to 'y'. Options `'y'` or `'x'`. Whether to make vertical
            or horizontal diamonds.
    Returns:
    ---------
        points (numpy.ndarray): the 2d array of points. 
    '''
    n_side = int(points_per_petal/2)
    axis_1 = np.concatenate((
                np.linspace(0, petal_width, int(n_side/2)), 
                np.linspace(petal_width, 0, int(n_side/2))
            ))
    axis_2 = np.linspace(0, 1, n_side)
    axes = (axis_1, axis_2) if direction == 'y' else (axis_2, axis_1)
    points = np.vstack(axes).T
    points = np.vstack((points, -1*points))
    points = np.vstack((points, np.vstack((points[:, 0], -1*points[:, 1])).T))
    return points

def make_diamonds(
    points_per_petal:int=200,
    petal_width:float=0.25,
    colors:int=5,
    scale_factor:float=30,
    use_gaussian:bool=True   
):
    '''
    Arguments:
    ----------
        points_per_petal (int). Defaults to `200`. Number of points per petal.
        petal_width (float): Defaults to `0.25`. How narrow the diamonds are.
        colors (int): Defaults to `5`. The number of timesteps (colors) to produce.
        scale_factor (float): Defaults to `30`. How much to scale the noise by 
            (larger values make samller noise).
        use_gaussian (bool): Defaults to `True`. Whether to use random or gaussian noise.
    Returns:
    ---------
        df (pandas.DataFrame): DataFrame with columns `samples`, `x`, `y`, where `samples`
            are the time index (corresponds to colors) 
    '''    
    upper = construct_diamond(points_per_petal, petal_width, 'y')
    lower = construct_diamond(points_per_petal, petal_width, 'x')
    data = np.vstack((upper, lower)) 
    
    noise_fn = np.random.randn if use_gaussian else np.random.rand
    noise = noise_fn(*data.shape) / scale_factor
    data = data + noise
    df = pd.DataFrame(data, columns=['d1', 'd2'])
    
    c_values = np.linspace(colors, 1, colors)
    c_thresholds = np.linspace(1, 0+1/(colors+1), colors)
    df.insert(0, 'samples', colors)
    df['samples'] = colors 
    for value, threshold in zip(c_values, c_thresholds):
        index = ((np.abs(df.d1) <= threshold) & (np.abs(df.d2) <= threshold))
        df.loc[index, 'samples'] = value
    df.set_index('samples')
    return df

# %% ../nbs/07_datasets.ipynb 6
def make_swiss_roll(n_points=1500):
    '''
    Arguments:
    ----------
        n_points (int): Default to `1500`. 

    Returns:
    ---------
        df (pandas.DataFrame): DataFrame with columns `samples`, `d1`, `d2`, `d3`, 
            where `samples` are the time index (corresponds to colors) 
    '''  
    X, color = datasets.make_swiss_roll(n_samples=n_points)
    df = pd.DataFrame(np.hstack((np.round(color).reshape(-1, 1), X)), columns='samples d1 d2 d3'.split())
    df.samples -= np.min(df.samples)
    return df

# %% ../nbs/07_datasets.ipynb 7
def make_tree():
    '''
    Arguments:
    ----------
        n_points (int): Default to `1500`. 
        
    Returns:
    ---------
        df (pandas.DataFrame): DataFrame with columns `samples`, `d1`, `d2`, `d3`, 
            `d4`, `d5` where `samples` are the time index (corresponds to colors) 
    '''  
    tree, branches = phate.tree.gen_dla(
        n_dim = 200, n_branch = 10, branch_length = 300, 
        rand_multiplier = 2, seed=37, sigma = 5
    )
    phate_operator = phate.PHATE(n_components=5, n_jobs=-1)
    tree_phate = phate_operator.fit_transform(tree)
    df = pd.DataFrame(np.hstack((branches.reshape(-1, 1), tree_phate)), columns='samples d1 d2 d3 d4 d5'.split())
    return df

# %% ../nbs/07_datasets.ipynb 8
from .constants import WORM_FILE
def make_worm_data():
    data = np.load(WORM_FILE)
    sample_labels = data['sample_labels']
    embedding = data['embedding']
    df = pd.concat([
            pd.DataFrame(sample_labels, columns=['samples']), 
            pd.DataFrame(embedding, columns=list(map(lambda e: f'd{e}', '12345')))
        ], axis=1,
    )
    df.set_index('samples')
    return df

# %% ../nbs/07_datasets.ipynb 9
from .constants import EB_BODIES_FILE,EB_BODIES_PSEUDO_4,EB_BODIES_PSEUDO_6,EB_BODIES_PSEUDO_25,EB_BODIES_PSEUDO_82

def make_eb_data(phate=False, phate_dims=5,n_sample='all', random_state=1):
    data = np.load(EB_BODIES_FILE)
    sample_labels = data['sample_labels']
    embedding = data['pca']
    df = pd.DataFrame(embedding, columns=[f'd{i}' for i in range(1, 101)])
    df['samples'] = sample_labels    
    df.set_index('samples')
    df['pt4'] = np.load(EB_BODIES_PSEUDO_4)
    df['pt6'] = np.load(EB_BODIES_PSEUDO_6)
    df['pt25'] = np.load(EB_BODIES_PSEUDO_25)
    df['pt82'] = np.load(EB_BODIES_PSEUDO_82)
    if n_sample != 'all' and not phate:
        df = df.sample(n=n_sample,random_state=random_state)
    
    if phate:
        from phate import PHATE
        phate_operator = PHATE(phate_dims, n_jobs=-1)
        sub_sample = df.sample(n=n_sample,random_state=random_state)
        Y_phate = phate_operator.fit_transform(sub_sample[[f'd{i}' for i in range(1, 11)]])
        df = pd.concat([
        pd.DataFrame(df.samples.values, columns=['samples']), 
        pd.DataFrame(Y_phate, columns=list(map(lambda e: f'd{e}', range(1, phate_dims+1))))
        , df['pt4'], df['pt6'], df['pt25']], axis=1)
    return df

# %% ../nbs/07_datasets.ipynb 10
from .constants import (DYNGEN_INFO_FILE, DYNGEN_EXPR_FILE)
from phate import PHATE
import warnings
def make_dyngen_data(
    time_col='sim_time', phate_dims=10, round_labels=True,
    use_gaussian:bool=False, add_noise=False, add_noise_after_phate=False,
    scale_factor:float=1, scale_phate=100, n_bins=5, column='d1'
):
    _valid = 'simulation_i step_ix sim_time'.split()
    if time_col not in _valid:
        time_col = _valid[0]        

    noise_fn = np.random.randn if use_gaussian else np.random.rand
    
    
    exp = pd.read_csv(DYNGEN_EXPR_FILE, )

    if add_noise and not add_noise_after_phate:
        noise = noise_fn(*exp.shape) / scale_factor
        exp += noise

    ids = pd.read_csv(DYNGEN_INFO_FILE, skipfooter=1, engine='python').dropna(axis=1)
    df = pd.concat([ids, exp], axis=1).set_index('cell_id')
    df['samples'] = df[time_col]
    df = df.drop(columns=_valid)

    phate_operator = PHATE(phate_dims, n_jobs=-1)
    Y_phate = phate_operator.fit_transform(df.drop(columns=['samples']))

    Y_phate *= scale_phate

    if add_noise and add_noise_after_phate:
        noise = noise_fn(*Y_phate.shape) / scale_factor
        Y_phate += noise

    df = pd.concat([
        pd.DataFrame(df.samples.values, columns=['samples']), 
        pd.DataFrame(Y_phate, columns=list(map(lambda e: f'd{e}', range(1, phate_dims+1))))
    ], axis=1)

    if round_labels:
        # instead of 0 - 1000 ----> 0 - 10
        df.samples = np.round(df.samples, -2) / 100

    df = relabel_data(df, min_bin=0, n_bins=n_bins, column=column)

    if phate_dims in [2,5]:
        locs = (df['d1'] <= -2.0)
        df.loc[locs, 'samples'] = -1
        df.drop(df[df['samples'] == -1].index, inplace = True)
    elif phate_dims in [10,15,30,40,60]:
        locs = (df['d1'] <= -1.9)
        df.loc[locs, 'samples'] = -1
        df.drop(df[df['samples'] == -1].index, inplace = True)
    else:
        warnings.warn('Not tested for this \'phate_dims\', using the same threshold as the one from \'[10,15,30,40,60]\' dims.')
        locs = (df['d1'] <= -1.9)
        df.loc[locs, 'samples'] = -1
        df.drop(df[df['samples'] == -1].index, inplace = True)

    return df



def relabel_data(df,min_bin=0, n_bins=10, column='d1', samples_key='samples'):
    dff = df.copy()
    
    x_min = np.min(dff[column])
    x_max = np.max(dff[column])
    
    parts = np.linspace(x_min, x_max, n_bins+1)
    value = list(range(min_bin, n_bins+1, 1))
    for i, x in list(zip(value, parts))[::-1]:
        if i == 0:
            continue
        locs = (dff[column] <= x)
        dff.loc[locs, samples_key] = i
    return dff

# %% ../nbs/07_datasets.ipynb 12
import numpy as np, seaborn as sns, pandas as pd, matplotlib.pyplot as plt

def rings(
    N:int, M:int = None, 
    data_scale:float = 1, 
    add_noise:bool = True, 
    noise_scale_theta:float = 0.7, 
    noise_scale_radius:float = 0.03,
    buffer:float = 0.8,
    **kwargs
) -> (np.ndarray, np.ndarray):
    '''
    Arguments:
        N (int): Number of points to make.
        M (int): Defaults to `None`. If `M='auto'` will automatically determine how many circles to make.
        data_scale (float): Defaults to `1`. Multiplier to rescale the data.
        add_noise (bool): Defaults to `True`. Whether or not to add noise to the data.
        noise_scale_theta (float): Defaults to `0.7`. How much to scale the noise added to `theta`.
        noise_scale_radius (float): Defaults to `0.3`. How much to scale the noise added to `radius`.
        buffer (float): Defaults to `0.8`. How much to scale the `radius` to add some padding between circles.
        **kwargs    
        
    Returns:
        X (np.ndarray): The x, y coordinates for the points.
        C (np.ndarray): The cluster number of each point.
    '''
    
    """Generate petal data set."""
    X = []  # points in respective petals
    Y = []  # auxiliary array (points on outer circle)
    C = []

    assert N > 4, "Require more than four data points"

    # Number of 'petals' to point into the data set. This is required to
    # ensure that the full space is used.
    if M is None:
        M = int(np.floor(np.sqrt(N)))
    thetas = np.linspace(0, 2 * np.pi, M, endpoint=False)
    

    for theta in thetas:
        Y.append(np.asarray([np.cos(theta), np.sin(theta)]))
        
    # Radius of the smaller cycles is half of the chord distance between
    # two 'consecutive' points on the circle.
    radius = 0.5 * np.linalg.norm(Y[0] - Y[1])    

    for i, x in enumerate(Y):
        for theta in thetas:
            for j in range(N // M // len(thetas)):
                r = radius if not add_noise else radius + np.random.randn() * noise_scale_radius
                t = theta if not add_noise else theta + np.random.randn() * noise_scale_theta
                r *= buffer
                X.append(np.asarray([r * np.cos(t) - x[0], r * np.sin(t) - x[1]]))

                # Indicates that this point belongs to the $i$th circle.
                C.append(i)
    X = np.asarray(X)
    C = np.asarray(C)
    X *= data_scale
    return X, C

def make_rings(N:int, M:int = None, 
    data_scale:float = 1, 
    add_noise:bool = True, 
    noise_scale_theta:float = 0.7, 
    noise_scale_radius:float = 0.03,
    buffer:float = 0.8,
    **kwargs
) -> pd.DataFrame:
    '''
    Arguments:
        N (int): Number of points to make.
        M (int): Defaults to `None`. If `M='auto'` will automatically determine how many circles to make.
        data_scale (float): Defaults to `1`. Multiplier to rescale the data.
        add_noise (bool): Defaults to `True`. Whether or not to add noise to the data.
        noise_scale_theta (float): Defaults to `0.7`. How much to scale the noise added to `theta`.
        noise_scale_radius (float): Defaults to `0.3`. How much to scale the noise added to `radius`.
        buffer (float): Defaults to `0.8`. How much to scale the `radius` to add some padding between circles.
        **kwargs    
        
    Returns:
        X (np.ndarray): The x, y coordinates for the points.
        C (np.ndarray): The cluster number of each point.
    '''
    x, c = rings(N, M, data_scale, add_noise, noise_scale_theta, noise_scale_radius, buffer)
    df = pd.DataFrame(x, columns=[f'd{i+1}' for i in range(x.shape[1])])
    df['samples'] = c
    df.set_index('samples')
    return df

# %% ../nbs/07_datasets.ipynb 14
def make_jacks(
    n_axes = 3,
    points = 1000,
    label_by = 'axis',
    n_classes = 3,
    use_neg = True,
    data_scale = 3,
    add_noise = True,
    noise_scale = 0.03,
):

    _valid_label_bys = 'axis coord'.split()

    if label_by not in _valid_label_bys:
        label_by = _valid_label_bys[0]

    results = []
    classes = []

    axes = np.eye(n_axes)

    for i, axis in enumerate(axes):
        segment = np.linspace(0, 1, points // n_axes).reshape(-1, 1)
        if add_noise:
            coordinates = axis * (segment + np.random.randn(segment.size, 1) * noise_scale)
        else:
            coordinates = axis * segment
        results.extend(coordinates.tolist())

        if label_by == 'axis':
            labels = [i for j in range(len(segment))]
            classes.extend(labels)
        elif label_by == 'coord':
            labels = [
                k for k in range(n_classes)
                for j in range(points // n_axes // n_classes)
            ]
            for j in range(len(segment) - len(labels)):
                labels.append(n_classes - 1)
            classes.extend(labels)

    if use_neg:
        for i, axis in enumerate(axes):
            segment = np.linspace(0, 1, points // n_axes).reshape(-1, 1) * -1
            if add_noise:
                coordinates = axis * (segment + np.random.randn(segment.size, 1) * noise_scale)
            else:
                coordinates = axis * segment
            results.extend(coordinates.tolist())

            if label_by == 'axis':
                labels = [n_axes + i  for j in range(len(segment))]
                classes.extend(labels)
            elif label_by == 'coord':
                labels = [
                    k for k in range(n_classes)
                    for j in range(points // n_axes // n_classes)
                ]
                for j in range(len(segment) - len(labels)):
                    labels.append(n_classes - 1)
                classes.extend(labels)

    results = np.array(results) + np.random.randn(len(results), n_axes) * noise_scale
    results *= data_scale
    df = pd.DataFrame(results, columns=[f'd{i+1}' for i in range(n_axes)])
    df['samples'] = classes
    df.set_index('samples')
    return df

# %% ../nbs/07_datasets.ipynb 16
def branch_data_clean(t, e):
    shapes = (len(e), len(t))
    t = t.reshape(1,-1)
    e = e.reshape(-1, 1)
    ts = np.tile(t, (shapes[0], 1))[...,None]
    es = np.tile(e, (1, shapes[1]))[...,None]
    x = np.tile(t, (shapes[0], 1))
    y = e * t**2
    data = np.stack([x,y], axis=2)
    data = np.concatenate((ts, es, data), axis=-1)
    data = data.reshape(shapes[0]*shapes[1], -1)
    return data
def branch_data_data(data, n_colors=5, repeats=5, noisex=0.05, noisey=0.05, seed=32):
    data = np.tile(data, (repeats,1))
    df = pd.DataFrame(data, columns=['t', 'e1', 'd1', 'd2'])
    np.random.seed(seed)
    df['d1'] += np.random.randn(df.shape[0]) * noisex
    df['d2'] += np.random.randn(df.shape[0]) * noisey
    _, bin_edges = np.histogram(df['t'], bins=n_colors)
    bin_indices = np.digitize(df['t'], bin_edges)
    bin_indices[bin_indices > n_colors] = n_colors
    df.insert(0, 'samples', bin_indices)
    df.drop(columns=['t'], inplace=True)
    df.samples -=1
    return df
def make_branch_cond(nt=20, ne=3, n_colors=5, repeats=10, noisex=0.1, noisey=0.1, seed=32):
    t = np.linspace(0,1,nt)
    e = np.linspace(-1,1,ne)
    data = branch_data_clean(t, e)
    df = branch_data_data(data, n_colors, repeats, noisex, noisey, seed)
    return df

# %% ../nbs/07_datasets.ipynb 18
def make_partial_branch(ne=2, noisex=0.05, noisey=.1, n_colors=5, repeats=20, drop_e=True):
    df = make_branch_cond(ne=ne, n_colors=n_colors, repeats=repeats, noisex=noisex, noisey=noisey)
    top_branch = (df['e1'] == 1) & (df['samples'].isin([2,3,4]))
    df.loc[top_branch, 'd2'] = (df[top_branch]['d2'] - df[top_branch]['d2'].mean())/3 + df[top_branch]['d2'].mean() - 0.2
    top_drop = np.random.choice(np.where(top_branch)[0], int(top_branch.sum()*2/3))
    df = df.drop(top_drop, axis=0)
    if drop_e:
        df.drop(columns=['e1'], inplace=True)
    return df

# %% ../nbs/07_datasets.ipynb 20
import numpy as np
import pandas as pd

def make_uniform_rect(x_bounds, y_bounds, num_points):
    """
    Generates uniformly distributed points within a rectangular box.

    Parameters:
    - x_bounds: Tuple (x_min, x_max) defining the bounds on the X-axis.
    - y_bounds: Tuple (y_min, y_max) defining the bounds on the Y-axis.
    - num_points: The number of points to generate.

    Returns:
    - points: A 2D numpy array where each row is a point [x, y].
    """
    x_min, x_max = x_bounds
    y_min, y_max = y_bounds
    
    x_points = np.random.uniform(x_min, x_max, num_points)
    y_points = np.random.uniform(y_min, y_max, num_points)

    points = np.column_stack((x_points, y_points))
    
    return points

import numpy as np
import pandas as pd

def make_gaussian_rect(x_bounds, y_bounds, num_points):
    """
    Generates Gaussian distributed points within a rectangular box.
    
    Parameters:
    - x_bounds: Tuple (x_min, x_max) defining the bounds on the X-axis.
    - y_bounds: Tuple (y_min, y_max) defining the bounds on the Y-axis.
    - num_points: The number of points to generate.
    
    Returns:
    - points: A 2D numpy array where each row is a point [x, y].
    """
    x_min, x_max = x_bounds
    y_min, y_max = y_bounds
    
    # Calculate the mean and standard deviation based on the bounds
    x_mean = (x_min + x_max) / 2
    y_mean = (y_min + y_max) / 2
    x_std = (x_max - x_min) / 3  # 99.7% of data within bounds for ±3 std
    y_std = (y_max - y_min) / 6
    
    # Generate Gaussian distributed points within the rectangle
    x_points = np.random.normal(x_mean, x_std, num_points)
    y_points = np.random.normal(y_mean, y_std, num_points)
    
    # Clip the points to stay within the specified bounds
    x_points = np.clip(x_points, x_min, x_max)
    y_points = np.clip(y_points, y_min, y_max)
    
    # Combine the points into a 2D array (each row is a point [x, y])
    points = np.column_stack((x_points, y_points))
    
    return points

def make_dying_example_unif(n_pts_per_bin=50, seed=223):
    np.random.seed(seed)
    tp0 = make_uniform_rect((-1.5, -0.4), (-0.5, 0.5), n_pts_per_bin*2)
    tp1 = make_uniform_rect((-0.6, 0.6), (-0.5, 0.5), n_pts_per_bin*2)
    tp2 = make_uniform_rect((0.4, 1.6), (-0.5, 0.5), n_pts_per_bin*2)
    tp3 = make_uniform_rect((1.4, 2.6), (0., 0.5), n_pts_per_bin)
    tp4 = make_uniform_rect((2.4, 3.1), (0., 0.5), n_pts_per_bin)
    # concatenate, but add a column for the timepoint, make it a dataframe
    # Create dataframes and add a timepoint column

    df_tp0 = pd.DataFrame(tp0, columns=['d1', 'd2'])
    df_tp0['samples'] = 0

    df_tp1 = pd.DataFrame(tp1, columns=['d1', 'd2'])
    df_tp1['samples'] = 1

    df_tp2 = pd.DataFrame(tp2, columns=['d1', 'd2'])
    df_tp2['samples'] = 2

    df_tp3 = pd.DataFrame(tp3, columns=['d1', 'd2'])
    df_tp3['samples'] = 3

    df_tp4 = pd.DataFrame(tp4, columns=['d1', 'd2'])
    df_tp4['samples'] = 4

    # Concatenate the dataframes
    df = pd.concat([df_tp0, df_tp1, df_tp2, df_tp3, df_tp4], ignore_index=True)
    return df

def make_dying_example_gaus(n_pts_per_bin=50, seed=223):
    np.random.seed(seed)
    tp0 = make_gaussian_rect((-1.5, -0.4), (-0.5, 0.5), n_pts_per_bin*2)
    tp1 = make_gaussian_rect((-0.6, 0.6), (-0.5, 0.5), n_pts_per_bin*2)
    tp2 = make_gaussian_rect((0.4, 1.6), (-0.5, 0.5), n_pts_per_bin*2)
    tp3 = make_gaussian_rect((1.4, 2.6), (0., 0.5), n_pts_per_bin)
    tp4 = make_gaussian_rect((2.4, 3.1), (0., 0.5), n_pts_per_bin)
    # concatenate, but add a column for the timepoint, make it a dataframe
    # Create dataframes and add a timepoint column

    df_tp0 = pd.DataFrame(tp0, columns=['d1', 'd2'])
    df_tp0['samples'] = 0

    df_tp1 = pd.DataFrame(tp1, columns=['d1', 'd2'])
    df_tp1['samples'] = 1

    df_tp2 = pd.DataFrame(tp2, columns=['d1', 'd2'])
    df_tp2['samples'] = 2

    df_tp3 = pd.DataFrame(tp3, columns=['d1', 'd2'])
    df_tp3['samples'] = 3

    df_tp4 = pd.DataFrame(tp4, columns=['d1', 'd2'])
    df_tp4['samples'] = 4

    # Concatenate the dataframes
    df = pd.concat([df_tp0, df_tp1, df_tp2, df_tp3, df_tp4], ignore_index=True)
    return df


# %% ../nbs/07_datasets.ipynb 22
import numpy as np
import pandas as pd

def make_growing(n_points_per_part=[50,100,150,200], noise_level=0.1, x_max=1):
    """
    Generates random points in the region between y=x^2, y=-x^2, and x=x_max, and adds noise to it.
    The points are split evenly into parts with numbers equal to len(n_points_per_part) by x.
    
    Parameters:
    n_points_per_part (list of int): List containing the number of points to generate in each part.
    noise_level (float): Standard deviation of the Gaussian noise to be added.
    x_max (float): The maximum range of x.
    
    Returns:
    pd.DataFrame: DataFrame containing the generated points with columns 'd1', 'd2', and 'samples'.
    """
    n_parts = len(n_points_per_part)
    x_ranges = [(i * x_max / n_parts, (i + 1) * x_max / n_parts) for i in range(n_parts)]
    
    points = []
    
    for i, n_points in enumerate(n_points_per_part):
        x = np.random.uniform(x_ranges[i][0], x_ranges[i][1], n_points)
        y = np.random.uniform(-x**2, x**2, n_points)
        noise_x = np.random.normal(0, noise_level, n_points)
        noise_y = np.random.normal(0, noise_level, n_points)
        part_points = np.vstack((x + noise_x, y + noise_y)).T
        samples = np.full(n_points, i)
        part_df = pd.DataFrame(part_points, columns=['d1', 'd2'])
        part_df['samples'] = samples
        points.append(part_df)
    
    return pd.concat(points, ignore_index=True)




# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/MIOFlow/',
                'doc_host': 'https://KrishnaswamyLab.github.io',
                'git_url': 'https://github.com/KrishnaswamyLab/MIOFlow/tree/main/',
                'lib_path': 'MIOFlow'},
  'syms': { 'MIOFlow.constants': {},
            'MIOFlow.datasets': { 'MIOFlow.datasets.branch_data_clean': ('datasets.html#branch_data_clean', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.branch_data_data': ('datasets.html#branch_data_data', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.construct_diamond': ('datasets.html#construct_diamond', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_branch_cond': ('datasets.html#make_branch_cond', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_diamonds': ('datasets.html#make_diamonds', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_dying_example_gaus': ( 'datasets.html#make_dying_example_gaus',
                                                                                'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_dying_example_unif': ( 'datasets.html#make_dying_example_unif',
                                                                                'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_dyngen_data': ('datasets.html#make_dyngen_data', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_eb_data': ('datasets.html#make_eb_data', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_gaussian_rect': ('datasets.html#make_gaussian_rect', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_growing': ('datasets.html#make_growing', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_jacks': ('datasets.html#make_jacks', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_partial_branch': ('datasets.html#make_partial_branch', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_rings': ('datasets.html#make_rings', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_swiss_roll': ('datasets.html#make_swiss_roll', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_tree': ('datasets.html#make_tree', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_uniform_rect': ('datasets.html#make_uniform_rect', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.make_worm_data': ('datasets.html#make_worm_data', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.relabel_data': ('datasets.html#relabel_data', 'MIOFlow/datasets.py'),
                                  'MIOFlow.datasets.rings': ('datasets.html#rings', 'MIOFlow/datasets.py')},
            'MIOFlow.eval': { 'MIOFlow.eval.calculate_nn': ('eval.html#calculate_nn', 'MIOFlow/eval.py'),
                              'MIOFlow.eval.generate_plot_data': ('eval.html#generate_plot_data', 'MIOFlow/eval.py'),
                              'MIOFlow.eval.generate_points': ('eval.html#generate_points', 'MIOFlow/eval.py'),
                              'MIOFlow.eval.generate_tjnet_trajectories': ('eval.html#generate_tjnet_trajectories', 'MIOFlow/eval.py'),
                              'MIOFlow.eval.generate_trajectories': ('eval.html#generate_trajectories', 'MIOFlow/eval.py'),
                              'MIOFlow.eval.get_cell_indexes': ('eval.html#get_cell_indexes', 'MIOFlow/eval.py'),
                              'MIOFlow.eval.get_points_from_trajectories': ('eval.html#get_points_from_trajectories', 'MIOFlow/eval.py')},
            'MIOFlow.exp': { 'MIOFlow.exp.config_exp_logger': ('exp.html#config_exp_logger', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.exp_log_filename': ('exp.html#exp_log_filename', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.exp_param_filename': ('exp.html#exp_param_filename', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.find_exps': ('exp.html#find_exps', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.gen_exp_name': ('exp.html#gen_exp_name', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.is_config_subset': ('exp.html#is_config_subset', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.list_exps': ('exp.html#list_exps', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.load_exp_params': ('exp.html#load_exp_params', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.save_exp_params': ('exp.html#save_exp_params', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.setup_exp': ('exp.html#setup_exp', 'MIOFlow/exp.py'),
                             'MIOFlow.exp.test_exp': ('exp.html#test_exp', 'MIOFlow/exp.py')},
            'MIOFlow.geo': { 'MIOFlow.geo.DiffusionAffinity': ('geo.html#diffusionaffinity', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionAffinity.__init__': ('geo.html#diffusionaffinity.__init__', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionAffinity.fit': ('geo.html#diffusionaffinity.fit', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionDistance': ('geo.html#diffusiondistance', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionDistance.__init__': ('geo.html#diffusiondistance.__init__', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionDistance.compute_custom_diffusion_distance': ( 'geo.html#diffusiondistance.compute_custom_diffusion_distance',
                                                                                                  'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionDistance.compute_stationnary_distrib': ( 'geo.html#diffusiondistance.compute_stationnary_distrib',
                                                                                            'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionDistance.fit': ('geo.html#diffusiondistance.fit', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionMap': ('geo.html#diffusionmap', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionMap.__init__': ('geo.html#diffusionmap.__init__', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.DiffusionMap.fit': ('geo.html#diffusionmap.fit', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.PhateDistance': ('geo.html#phatedistance', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.PhateDistance.__init__': ('geo.html#phatedistance.__init__', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.PhateDistance.fit': ('geo.html#phatedistance.fit', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance': ('geo.html#old_diffusiondistance', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.__init__': ('geo.html#old_diffusiondistance.__init__', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.compute_custom_diffusion_distance': ( 'geo.html#old_diffusiondistance.compute_custom_diffusion_distance',
                                                                                                      'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.compute_density_norm_matrix': ( 'geo.html#old_diffusiondistance.compute_density_norm_matrix',
                                                                                                'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.compute_diffusion_Matrix': ( 'geo.html#old_diffusiondistance.compute_diffusion_matrix',
                                                                                             'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.compute_stationnary_distrib': ( 'geo.html#old_diffusiondistance.compute_stationnary_distrib',
                                                                                                'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.distance_matrix_Pt': ( 'geo.html#old_diffusiondistance.distance_matrix_pt',
                                                                                       'MIOFlow/geo.py'),
                             'MIOFlow.geo.old_DiffusionDistance.fit': ('geo.html#old_diffusiondistance.fit', 'MIOFlow/geo.py'),
                             'MIOFlow.geo.setup_distance': ('geo.html#setup_distance', 'MIOFlow/geo.py')},
            'MIOFlow.losses': { 'MIOFlow.losses.Density_loss': ('losses.html#density_loss', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.Density_loss.__call__': ('losses.html#density_loss.__call__', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.Density_loss.__init__': ('losses.html#density_loss.__init__', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.Local_density_loss': ('losses.html#local_density_loss', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.Local_density_loss.__call__': ( 'losses.html#local_density_loss.__call__',
                                                                                'MIOFlow/losses.py'),
                                'MIOFlow.losses.Local_density_loss.__init__': ( 'losses.html#local_density_loss.__init__',
                                                                                'MIOFlow/losses.py'),
                                'MIOFlow.losses.MMD_loss': ('losses.html#mmd_loss', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.MMD_loss.__init__': ('losses.html#mmd_loss.__init__', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.MMD_loss.forward': ('losses.html#mmd_loss.forward', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.MMD_loss.guassian_kernel': ('losses.html#mmd_loss.guassian_kernel', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.OT_loss': ('losses.html#ot_loss', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.OT_loss.__call__': ('losses.html#ot_loss.__call__', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.OT_loss.__init__': ('losses.html#ot_loss.__init__', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.covariance_loss': ('losses.html#covariance_loss', 'MIOFlow/losses.py'),
                                'MIOFlow.losses.ot_loss_given_plan': ('losses.html#ot_loss_given_plan', 'MIOFlow/losses.py')},
            'MIOFlow.models': { 'MIOFlow.models.Autoencoder': ('models.html#autoencoder', 'MIOFlow/models.py'),
                                'MIOFlow.models.Autoencoder.__init__': ('models.html#autoencoder.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.Autoencoder.decode': ('models.html#autoencoder.decode', 'MIOFlow/models.py'),
                                'MIOFlow.models.Autoencoder.encode': ('models.html#autoencoder.encode', 'MIOFlow/models.py'),
                                'MIOFlow.models.Autoencoder.forward': ('models.html#autoencoder.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalModel': ('models.html#conditionalmodel', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalModel.__init__': ('models.html#conditionalmodel.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalModel.eval': ('models.html#conditionalmodel.eval', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalModel.forward': ('models.html#conditionalmodel.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalModel.train': ('models.html#conditionalmodel.train', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalODE': ('models.html#conditionalode', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalODE.__init__': ('models.html#conditionalode.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalODE.forward': ('models.html#conditionalode.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalODE.set_condition': ( 'models.html#conditionalode.set_condition',
                                                                                 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalSDEModel': ('models.html#conditionalsdemodel', 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalSDEModel.__init__': ( 'models.html#conditionalsdemodel.__init__',
                                                                                 'MIOFlow/models.py'),
                                'MIOFlow.models.ConditionalSDEModel.forward': ( 'models.html#conditionalsdemodel.forward',
                                                                                'MIOFlow/models.py'),
                                'MIOFlow.models.GrowthRateModel': ('models.html#growthratemodel', 'MIOFlow/models.py'),
                                'MIOFlow.models.GrowthRateModel.__init__': ('models.html#growthratemodel.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.GrowthRateModel.forward': ('models.html#growthratemodel.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyModel': ('models.html#toymodel', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyModel.__init__': ('models.html#toymodel.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyModel.eval': ('models.html#toymodel.eval', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyModel.forward': ('models.html#toymodel.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyModel.train': ('models.html#toymodel.train', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyODE': ('models.html#toyode', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyODE.__init__': ('models.html#toyode.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToyODE.forward': ('models.html#toyode.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel': ('models.html#toysdemodel', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel.__init__': ('models.html#toysdemodel.__init__', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel.eval': ('models.html#toysdemodel.eval', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel.f': ('models.html#toysdemodel.f', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel.forward': ('models.html#toysdemodel.forward', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel.g': ('models.html#toysdemodel.g', 'MIOFlow/models.py'),
                                'MIOFlow.models.ToySDEModel.train': ('models.html#toysdemodel.train', 'MIOFlow/models.py'),
                                'MIOFlow.models.make_model': ('models.html#make_model', 'MIOFlow/models.py')},
            'MIOFlow.ode': { 'MIOFlow.ode.NeuralODE': ('ode.html#neuralode', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.NeuralODE.__init__': ('ode.html#neuralode.__init__', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.NeuralODE.forward': ('ode.html#neuralode.forward', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ODEAdjoint': ('ode.html#odeadjoint', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ODEAdjoint.backward': ('ode.html#odeadjoint.backward', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ODEAdjoint.forward': ('ode.html#odeadjoint.forward', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ODEF': ('ode.html#odef', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ODEF.flatten_parameters': ('ode.html#odef.flatten_parameters', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ODEF.forward_with_grad': ('ode.html#odef.forward_with_grad', 'MIOFlow/ode.py'),
                             'MIOFlow.ode.ode_solve': ('ode.html#ode_solve', 'MIOFlow/ode.py')},
            'MIOFlow.plots': { 'MIOFlow.plots.new_plot_comparisions': ('plots.html#new_plot_comparisions', 'MIOFlow/plots.py'),
                               'MIOFlow.plots.plot_comparision': ('plots.html#plot_comparision', 'MIOFlow/plots.py'),
                               'MIOFlow.plots.plot_gene_trends': ('plots.html#plot_gene_trends', 'MIOFlow/plots.py'),
                               'MIOFlow.plots.plot_losses': ('plots.html#plot_losses', 'MIOFlow/plots.py')},
            'MIOFlow.train': { 'MIOFlow.train.train': ('train.html#train', 'MIOFlow/train.py'),
                               'MIOFlow.train.train_ae': ('train.html#train_ae', 'MIOFlow/train.py'),
                               'MIOFlow.train.training_regimen': ('train.html#training_regimen', 'MIOFlow/train.py')},
            'MIOFlow.utils': { 'MIOFlow.utils.config_criterion': ('utils.html#config_criterion', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.config_hold_out': ('utils.html#config_hold_out', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.generate_steps': ('utils.html#generate_steps', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.get_cell_types_from_df': ('utils.html#get_cell_types_from_df', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.get_groups_from_df': ('utils.html#get_groups_from_df', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.get_sample_n_from_df': ('utils.html#get_sample_n_from_df', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.get_times_from_groups': ('utils.html#get_times_from_groups', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.group_extract': ('utils.html#group_extract', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.sample': ('utils.html#sample', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.set_seeds': ('utils.html#set_seeds', 'MIOFlow/utils.py'),
                               'MIOFlow.utils.to_np': ('utils.html#to_np', 'MIOFlow/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/09_geo.ipynb.

# %% auto 0
__all__ = ['DiffusionDistance', 'DiffusionAffinity', 'DiffusionMap', 'PhateDistance', 'old_DiffusionDistance', 'setup_distance']

# %% ../nbs/09_geo.ipynb 3
import graphtools
from scipy.sparse import csr_matrix
import numpy as np
from sklearn.metrics.pairwise import pairwise_distances

class DiffusionDistance:
    """
    class DiffusionDistance        
        X (np.array) data 
        t_max (int), 2^t_max is the max scale of the Diffusion kernel
        knn (int) = 5 number of neighbors for the KNN in the alpha decay kernel construction, same default as in PHATE
        Anisotropy (int): the alpha in Coifman Lafon 2006, 1: double normalization 0: usual random walk
        log (bool) log(P) or not
        normalize (bool) min-max normalization of the distance matrix
        phate (bool) is PHATE op if true (should be the same as graphtool)
        
    """
   
    def __init__(self, t_max=5, knn=5, anisotropy=1, log=False, normalize=False, symmetrize=False) -> None:
        self.t_max = t_max
        self.knn = knn
        self.aniso = anisotropy
        self.log = log
        self.normalize = normalize
        self.symmetrize = symmetrize
        self.K = None
        self.P = None
        self.pi = None
        self.G = None    
            
    def compute_stationnary_distrib(self): 
        pi = np.sum(self.K, axis = 1)
        self.pi = (pi/np.sum(pi)).reshape(-1,1)
        return self.pi
        
    def compute_custom_diffusion_distance(self): 
        P = self.P
        P_d = P if not self.log else csr_matrix((np.log(P.data),P.indices,P.indptr), shape=P.shape)
        G = pairwise_distances(P_d,P_d,metric='l1',n_jobs=-1)
                
        for t in range(1,self.t_max): 
            P = P @ P 
            if self.log==True:
                dist = pairwise_distances(P,P,metric='l1',n_jobs=-1)
                np.fill_diagonal(dist,1)
                dist = (-1)*np.log(dist)
            else:
                dist = pairwise_distances(P_d,P_d,metric='l1',n_jobs=-1)
            G = G + 2**(-t/2.0) * dist
        
        if self.log==True:
            dist = pairwise_distances(self.pi,self.pi,metric='l1',n_jobs=-1)
            np.fill_diagonal(dist,1)
            dist = (-1)*np.log(dist)
        else:
            dist = pairwise_distances(self.pi,self.pi,metric='l1',n_jobs=-1)     
        G = G + 2**(-(self.t_max+1)/2.0) * dist
        self.G = G if not self.normalize else (G - np.min(G))/(np.max(G)-np.min(G))
        return self.G

    def fit(self, X):
        graph = graphtools.Graph(X, knn=self.knn,anisotropy=self.aniso)
        self.K = graph.K.toarray()
        self.P = graph.diff_op.toarray() 
        self.compute_stationnary_distrib()
        self.compute_custom_diffusion_distance()       
        return self.G if not self.symmetrize else (self.G + np.transpose(self.G))/0.5

# %% ../nbs/09_geo.ipynb 4
import phate
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import eigs
import sklearn
import graphtools

class DiffusionAffinity:
    """
    class DiffusionAffinity        
        X (np.array) data 
        t_max (int), 2^t_max is the max scale of the Diffusion kernel
        knn (int) = 5 number of neighbors for the KNN in the alpha decay kernel construction, same default as in PHATE
        Anisotropy (int): the alpha in Coifman Lafon 2006, 1: double normalization 0: usual random walk
        t_diff (int) the power of the diffusion affinity matrix
        topeig (int) in the the top k eigenvalues to consider in the spectral decomposition
        
        return the l2 between the row of the affinity matrix A^t
    """
   
    def __init__(self, knn=5, anisotropy=0, t_diff=1, topeig=100) -> None:
        self.knn = knn
        self.aniso = anisotropy
        self.t_diff = t_diff
        self.topeig = topeig
        self.A = None
        self.G = None    
        
    def fit(self, X):
        self.graph = graphtools.Graph(X, knn=self.knn,anisotropy=self.aniso) 
        self.A = self.graph.diff_aff
        if self.t_diff ==1:
            self.G = sklearn.metrics.pairwise.pairwise_distances(self.A,self.A,metric='l2',n_jobs=-1)
        else:
            k = self.topeig if self.topeig < X.shape[0] else X.shape[0]-2
            w, v = eigs(self.A,k=k, which='LR')
            W = np.diag(w)
            A_t = np.real(v @ (W**self.t_diff) @ v.T)
            self.G = sklearn.metrics.pairwise.pairwise_distances(A_t,A_t,metric='l2',n_jobs=-1)
        return self.G

# %% ../nbs/09_geo.ipynb 5
import scipy
import graphtools

class DiffusionMap:
    """
    Arguments
    ---------     
        X (np.array) data 
        t_max (int), 2^t_max is the max scale of the Diffusion kernel
        knn (int) = 5 number of neighbors for the KNN in the alpha decay kernel construction, same default as in PHATE
        Anisotropy (int): the alpha in Coifman Lafon 2006, 1: double normalization 0: usual random walk
        t_diff (int) the power of the diffusion affinity matrix
        topeig (int) in the the top k eigenvalues to consider in the spectral decomposition
        n_emb (int) the dimension of the emb space
        
    Returns
    ------- 
        np.array 
            The pairwise dist. in the diffusion map embedding.
    """
   
    def __init__(self, knn=5, anisotropy=0, t_diff=1, topeig=100, n_emb=10) -> None:
        self.knn = knn
        self.aniso = anisotropy
        self.t_diff = t_diff
        self.topeig = topeig
        self.n_emb = n_emb
        self.A = None
        self.G = None    
        
    def fit(self, X):
        self.graph = graphtools.Graph(X, knn=self.knn,anisotropy=self.aniso) 
        self.A = self.graph.diff_aff
        k = self.topeig if self.topeig < X.shape[0] else X.shape[0]-2
        w, v = eigs(self.A,k=k, which='LR')
        w, v = np.real(w), np.real(v)
        v = v/v[:,0,None]
        self.emb = np.vstack([(w[i]**self.t_diff)*v[:,i] for i in range(1,self.n_emb+1)])
        self.G = scipy.spatial.distance.pdist(self.emb.T)
        return self.G

# %% ../nbs/09_geo.ipynb 6
from scipy.spatial.distance import pdist, squareform
import phate

class PhateDistance:
    """
    class
    -----
        Arguments
        ---------
        X (np.array) data 
        knn (int) = 5 number of neighbors for the KNN in the alpha decay kernel construction, same default as in PHATE
        Anisotropy (int): the alpha in Coifman Lafon 2006, 1: double normalization 0: usual random walk
        verbose (bool): verbose param. in PHATE.
    """
   
    def __init__(self, knn=5, anisotropy=0,verbose=False) -> None:
        self.knn = knn
        self.aniso = anisotropy
        self.verbose = verbose
        
    def fit(self, X # Dataset to fit.
            ):
        """
        Parameters
        ----------
            X: (np.array) Dataset to fit. 
        
        Returns
        -------
        np.array
            the L2 between Potential of Heat-diffusion
        """
        graph = phate.PHATE(knn=self.knn, verbose=self.verbose, n_landmark=X.shape[0]).fit(X)
        diff_pot = graph.diff_potential 
        self.G = squareform(pdist(diff_pot))
        return self.G

# %% ../nbs/09_geo.ipynb 8
import numpy as np
from scipy.spatial import distance_matrix

"""
class DiffusionDistance
    
    kernel
    X
    t_max
    stationnary_distrib
"""
class old_DiffusionDistance:
    kernel = None
    X = None
    t_max = None
    M = None
    P = None
    pi = None
    G = None
    def __init__(self, kernel, t_max) -> None:
        self.kernel= kernel
        self.t_max = t_max
     
    def compute_density_norm_matrix(self):
        K = self.kernel(self.X)
        Q = np.diag(np.sum(K, axis= 1))
        self.M = np.linalg.inv(Q).dot(K).dot(np.linalg.inv(Q))
        return self.M
    
    def compute_diffusion_Matrix(self): 
        D = np.diag(np.sum(self.M, axis= 1))
        self.P = np.linalg.inv(D).dot(self.M)
        return self.P
    
    def compute_stationnary_distrib(self): 
        pi = np.sum(self.M, axis = 1)/np.sum(self.M)
        self.pi = pi
        return self.pi
    
    def distance_matrix_Pt(self, t): 
        Pt = np.linalg.matrix_power(self.P, 2**(self.t_max-t))
        return distance_matrix(Pt,Pt,1)
        
    def compute_custom_diffusion_distance(self): 
        G = np.zeros((self.X.shape[0], self.X.shape[0]))
                
        for t in range(0,self.t_max): 
            G = G + 2**(-(self.t_max-t)/2) * self.distance_matrix_Pt(t)
        G = G + 2**(-(self.t_max+1)/2) * distance_matrix(self.pi[:, None],self.pi[:, None],1)

        self.G = G
        return self.G

    def fit(self, X):
        self.X = X
        self.compute_density_norm_matrix()
        self.compute_diffusion_Matrix()
        self.compute_stationnary_distrib()
        self.compute_custom_diffusion_distance()
        return self.G

# %% ../nbs/09_geo.ipynb 9
from sklearn.gaussian_process.kernels import RBF

def setup_distance(
    distance_type:str='gaussian',
    rbf_length_scale:float=0.5, t_max:int=5, knn:int=5, anisotropy:int=1,
):
    _valid_distance_types = 'gaussian alpha_decay phate'.split()
    
    if distance_type == 'gaussian':
        # TODO: rename / retool old_DiffusionDistance into single
        #      DiffusionDistance class that "automagically" figures out
        #      implementation via input params
        dist = old_DiffusionDistance(RBF(rbf_length_scale), t_max=t_max)
    elif distance_type == 'alpha_decay':
        dist = DiffusionDistance(knn=knn, t_max=t_max)

    elif distance_type == 'phate':
        dist = PhateDistance(knn=knn, anisotropy=anisotropy)

    else:
        raise NotImplementedError(
            f'distance_type={distance_type} is not an implemented distance.\n'
            f'Please see {_valid_distance_types}'
        )
    return dist



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_eval.ipynb.

# %% auto 0
__all__ = ['generate_points', 'generate_trajectories', 'generate_plot_data', 'get_points_from_trajectories', 'calculate_nn',
           'generate_tjnet_trajectories', 'get_cell_indexes']

# %% ../nbs/10_eval.ipynb 3
import torch, numpy as np
from .utils import sample, to_np

def generate_points(
    model, df, n_points=100, 
    sample_with_replacement=False, use_cuda=False, 
    samples_key='samples', sample_time=None, autoencoder=None, recon=False
):
    '''
    Arguments:
    ----------
        model (torch.nn.Module): Trained network with the property `ode` corresponding to a `NeuralODE(ODEF())`.
            See `MIOFlow.ode` for more.
        df (pd.DataFrame): DataFrame containing a column for the timepoint samples and the rest of the data.
        n_points (int): Number of points to generate.
        sample_with_replacement (bool): Defaults to `False`. Whether or not to use replacement when sampling
            initial timepoint.
        use_cuda (bool): Defaults to `False`. Whether or not to use cuda.
        samples_key (str): Defaults to `'samples'`. The column in the `df` which has the timepoint groups.
        sample_time (list | None): Defaults to `None`. If `None` uses the group numbers in order as the 
            timepoints as specified in the column `df[samples_key]`.
        autoencoder (nn.Module|NoneType): Default to None, the trained autoencoder.
        recon (bool): Default to 'False', whether to use the autoencoder for reconstruction.
    Returns:
    ----------
        generated (float[float[]]): a list with shape `(len(sample_time), n_points, len(df.columns) - 1)`
            of the generated points.
    '''
    to_torch = True #if use_cuda else False

    groups = sorted(df[samples_key].unique())
    if sample_time is None:
        sample_time = groups
    data_t0 = sample(
        df, np.min(groups), size=(n_points, ), 
        replace=sample_with_replacement, to_torch=to_torch, use_cuda=use_cuda
    )
    if autoencoder is not None and recon:
        data_t0 = torch.Tensor(data_t0).float()
        data_t0 = autoencoder.encoder(data_t0)
        
    time =  torch.Tensor(sample_time).cuda() if use_cuda else torch.Tensor(sample_time)
    generated = model(data_t0, time, return_whole_sequence=True)
    if autoencoder is not None and recon:
        generated = autoencoder.decoder(generated)
    return to_np(generated)

def generate_trajectories(
    model, df, n_trajectories=30, n_bins=100, 
    sample_with_replacement=False, use_cuda=False, samples_key='samples',autoencoder=None, recon=False
):
    '''
    Arguments:
    ----------
        model (torch.nn.Module): Trained network with the property `ode` corresponding to a `NeuralODE(ODEF())`.
            See `MIOFlow.ode` for more.
        df (pd.DataFrame): DataFrame containing a column for the timepoint samples and the rest of the data.
        n_trajectories (int): Number of trajectories to generate.
        n_bins (int): Number of bins to use for the trajectories. More makes it smoother. Defaults to `100`.
        sample_with_replacement (bool): Defaults to `False`. Whether or not to use replacement when sampling
            initial timepoint.
        use_cuda (bool): Defaults to `False`. Whether or not to use cuda.
        samples_key (str): Defaults to `'samples'`. The column in the `df` which has the timepoint groups.
        autoencoder (nn.Module|NoneType): Default to None, the trained autoencoder.
        recon (bool): Default to 'False', whether to use the autoencoder for reconstruction.
    Returns:
    ----------
        trajectories (float[float[]]): a list with shape `(n_bins, n_points, len(df.columns) - 1)`
            of the generated trajectories.
    '''
    groups = sorted(df[samples_key].unique())
    sample_time = np.linspace(np.min(groups), np.max(groups), n_bins)
    trajectories = generate_points(model, df, n_trajectories, sample_with_replacement, use_cuda, samples_key, sample_time,autoencoder=autoencoder, recon=recon)
    return trajectories
    
def generate_plot_data(
    model, df, n_points, n_trajectories, n_bins, 
    sample_with_replacement=False, use_cuda=False, samples_key='samples',
    logger=None, autoencoder=None, recon=False
):
    '''
    Arguments:
    ----------
        model (torch.nn.Module): Trained network with the property `ode` corresponding to a `NeuralODE(ODEF())`.
            See `MIOFlow.ode` for more.
        df (pd.DataFrame): DataFrame containing a column for the timepoint samples and the rest of the data.
        n_points (int): Number of points to generate.
        n_trajectories (int): Number of trajectories to generate.
        n_bins (int): Number of bins to use for the trajectories. More makes it smoother. Defaults to `100`.
        sample_with_replacement (bool): Defaults to `False`. Whether or not to use replacement when sampling
            initial timepoint.
        use_cuda (bool): Defaults to `False`. Whether or not to use cuda.
        samples_key (str): Defaults to `'samples'`. The column in the `df` which has the timepoint groups.
        autoencoder (nn.Module|NoneType): Default to None, the trained autoencoder.
        recon (bool): Default to 'False', whether to use the autoencoder for reconstruction.
    Returns:
    ----------
        points (float[float[]]): a list with shape `(len(df[sample_key].unique()), n_points, len(df.columns) - 1)`
            of the generated points.
        trajectories (float[float[]]): a list with shape `(n_bins, n_points, len(df.columns) - 1)`
            of the generated trajectories.
    '''
    if logger: logger.info(f'Generating points')
    points = generate_points(model, df, n_points, sample_with_replacement, use_cuda, samples_key, None, autoencoder=autoencoder, recon=recon)
    if logger: logger.info(f'Generating trajectories')
    trajectories = generate_trajectories(model, df, n_trajectories, n_bins, sample_with_replacement, use_cuda, samples_key, autoencoder=autoencoder, recon=recon)
    return points, trajectories

# %% ../nbs/10_eval.ipynb 4
import os, logging, sklearn, pandas as pd, numpy as np
from typing import Union
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal
from .utils import generate_steps

def get_points_from_trajectories(
    n_groups:int, 
    trajectories:Union[np.ndarray, list], 
    how:Union[Literal['start'], Literal['middle'], Literal['end']]='start', 
    logger:logging.Logger=None
) -> np.ndarray:
    '''
    Arguments:
        n_groups (int): how many time points there are total. 
        trajectories (np.ndarray | list): A list with three dimensions that correspond to:
            `(n_bins, n_points, n_dims)`, where `n_points` are the number of points / trajectories there are,
            `n_bins` correponds to how the trajectories were smoothed via binning (e.g. if there were 5 total
            time points one might have 100 bins to draw a smoother line when plotting), and `n_dims` are the
            number of dimensions the points are in (e.g. 2).
        how (str): Defaults to `'start'`. How to extract the point for the binned trajectories.
            If `'start'` takes the first point in the time window. If `'middle'` takes the floored averaged.
            If `'end'` takes the the last point in the time window.
        logger (logging.Logger): Defaults to `None`.
    Returns:
        points (np.ndarray): Points at the corresponding indicices. If `trajectories` has shape:
            `(n_bins, n_points, n_dims)`, this will have shape `(n_groups, n_points, n_dims)`.
    '''
    # Value handling
    _valid_how = 'start middle end'.split()
    if how not in _valid_how:
        raise ValueError(f'Unknown option specified for `how`. Must be in {_valid_how}')    
        
    if not isinstance(trajectories, np.ndarray):
        trajectories = np.array(trajectories)
        
    trajectories = np.transpose(trajectories, axes=(1, 0, 2))
    (n_points, n_bins, n_dims) = trajectories.shape
    if logger: 
        logger.info(
            f'Given trajectories object with {n_points} points of {n_bins} '
            f'bins in {n_dims} dimensions.'
        )
    
    parts = np.linspace(0, n_bins, n_groups + 1).astype(int).tolist()
    steps = generate_steps(parts)
    results = []
    for step in steps:
        time_window = trajectories[:, slice(*step)] 
        if how == 'start':
            results.append(time_window[:, 0, :])
        elif how == 'middle':
            idx = int(np.floor(n_bins / n_groups / 2))
            results.append(time_window[:, idx, :])
        elif how == 'end':
            results.append(time_window[:, -1, :])
        else:
            raise NotImplementedError(f'how={how} not implemented')        
    return np.array(results)


def calculate_nn(
    df:pd.DataFrame,
    generated:Union[np.ndarray, list]=None,
    trajectories:Union[np.ndarray, list]=None,
    compare_to:Union[Literal['time'], Literal['any']]='time',
    how:Union[Literal['start'], Literal['middle'], Literal['end']]='start',     
    k:int=1,
    groups:Union[np.ndarray, list]=None,
    sample_key:str='samples',
    method:str='mean',
    logger:logging.Logger=None
) -> float:
    '''
    Arguments:
        df (pd.DataFrame): DataFrame containing all points and time sample specified by `sample_key`
        generated (np.ndarray | list): A list of the generate points with shape 
            `(n_groups, n_points, n_dims)`, where `n_groups` is the total number of time indicies
            as specified in `groups`.
        trajectories (np.ndarray | list): A list with three dimensions that correspond to:
            `(n_bins, n_points, n_dims)`, where `n_points` are the number of points / trajectories there are,
            `n_bins` correponds to how the trajectories were smoothed via binning (e.g. if there were 5 total
            time points one might have 100 bins to draw a smoother line when plotting), and `n_dims` are the
            number of dimensions the points are in (e.g. 2).
        compare_to (str): Defaults to `'time'`. Determines points to use for KNN. If `'time'` will only 
            consider points at the same time index. If `'any'` will search for nearest points regardless
            of time.
        how (str): Defaults to `'start'`. How to extract the point for the binned trajectories.
            If `'start'` takes the first point in the time window. If `'middle'` takes the floored averaged.
            If `'end'` takes the the last point in the time window.
        k (int): Defaults to `1`. Number of points to compare predicted points in `generated` 
            or `trajectories` to.
        groups (np.ndarray | list): Defaults to `None` and will be extracted from `df` is not provided. 
            The sorted unique values of time samples from `df` as specified by `sample_key`.
        sample_key (str): Defaults to `'samples'`. The column in `df` which corresponds to the time index.
        method (str): Defaults to `'mean'`. If `'mean'` returns the mean of the knn distances. If `'quartile'`
            returns the mean of the worst (highest distances) quartile.
        logger (logging.Logger): Defaults to `None`.
    Returns:
        mean_dist (float): mean distance of predicted points to the `n` nearest-neighbor points.
    '''
    _valid_compare_to = 'time any'.split()
    if compare_to not in _valid_compare_to:
        raise ValueError(f'Unknown option specified for `compare_to`. Must be in {_valid_compare_to}') 
        
    _valid_how = 'start middle end'.split()
    if how not in _valid_how:
        raise ValueError(f'Unknown option specified for `how`. Must be in {_valid_how}')

    _valid_method = 'mean quartile'.split()
    if method not in _valid_method:
        raise ValueError(f'Unknown option specified for `method`. Must be in {_valid_method}')
        
    if trajectories is None and generated is None:
        raise ValueError(f'Either generated or trajectories must not be None!')
        
    if groups is None:
        groups = sorted(df[sample_key].unique())
    
    if generated is None:
        generated = get_points_from_trajectories(len(groups), trajectories, how, logger)
    
    distances = []    
    for idx, time_sample in enumerate(sorted(groups)):            
        pred_points = generated[idx]
        
        # NOTE: compare to points only at same time index
        if compare_to == 'time':
            true_points = df.groupby(sample_key).get_group(time_sample).drop(columns=sample_key).values
        # NOTE: compare to any point
        elif compare_to == 'any':
            true_points = df.drop(columns=sample_key).values
        else:            
            raise NotImplementedError(f'compare_to={compare_to} not implemented')
        true_points = true_points[:, :pred_points.shape[1]]
        neigh = sklearn.neighbors.NearestNeighbors(n_neighbors=k)
        neigh.fit(true_points)
        dists, indicies = neigh.kneighbors(pred_points, return_distance=True)
        distances.extend(dists.flatten().tolist())
    
    distances = np.array(distances)
    if method == 'mean':      
        return distances.mean()
    elif method == 'quartile':
        q1 = np.quantile(distances, 0.25)
        q2 = np.quantile(distances, 0.50)
        q3 = np.quantile(distances, 0.75)
        
        b1 = distances[np.where(distances < q1)]
        b2 = distances[np.where((distances < q2) & (distances >= q1))]
        b3 = distances[np.where((distances < q3) & (distances >= q2))]
        b4 = distances[np.where(distances >= q3)]

        return np.max([np.mean(b) for b in [b1, b2, b3, b4]])
    else:
        raise NotImplementedError(f'method={method} not implemented')

# %% ../nbs/10_eval.ipynb 5
from MIOFlow.utils import (
    to_np, get_groups_from_df, get_cell_types_from_df, 
    get_sample_n_from_df, get_times_from_groups
)
import seaborn as sns
def generate_tjnet_trajectories(
    model, df, n_bins=10, use_cuda=False, samples_key='samples', 
    autoencoder=None, recon=False, where='end', start=0
):
    '''
    Arguments:
    -----------
        model (nn.Module): Trained MIOFlow model.

        df (pd.DataFrame): DataFrame of shape (n_cells, dimensions + 1), where the extra column
            stems from a samples column (column indicating the timepoint of the cell). 
            By default the samples column is assumed to be `"samples"`.
        
        n_bins (int): For each time point split it into `n_bins` for smoother trajectories.
            If there are `t` time points then there will be `t * n_bins` total points.
        
        use_cuda (bool): Whether or not to use cuda for the model and autoencoder.
        
        samples_key (str): The name of the column in the `df` that corresponds to the time
            samples. Defaults to `"samples"`. 
        
        autoencoder (nn.Module) Trained Geodesic Autoencoder.
        
        recon (bool): Whether or not to use the `autoencoder` to reconstruct the output
            space from the `model`.
        
        where (str): Choices are `"start"`, and `"end"`. Defaults to `"end"`. Whether or not
            to start the trajectories at `t_0` (`"start"`) or `t_n` (`"end"`). 
    
        start (int): Defaults to `0`. Where in `generate_tjnet_trajectories` the trajectories started.
            This is used if attempting to generate outside of `t0`. Note this works relative to `where`.
            E.g. if `where="end"` and `start=0` then this is the same as `groups[-1]`.

    Returns:
    -----------
        trajectories (np.ndarray): Trajectories with shape (time, cells, dimensions)
    '''
    
    _valid_where = 'start end'.split()
    if where not in _valid_where:
        raise ValueError(f'{where} not known. Should be one of {_valid_where}')
    
    groups = sorted(df[samples_key].unique())
    
    # times = groups
    # if where == 'end':
    #     times = times[::-1]
    # times = times[start:]
    
    times = get_times_from_groups(groups, where, start)

    # a, b = (np.max(groups), np.min(groups)) if where == 'end' else (np.min(groups), np.max(groups))    
    a, b = (np.max(times), np.min(times)) if where == 'end' else (np.min(times), np.max(times))    
    # n = -1 if where == 'end' else 0
    n = 0 # because reversed if needed and prunned
    
    # sample_time = np.linspace(a, b, len(groups) * n_bins)
    sample_time = np.linspace(a, b, len(times) * n_bins)
    sample_time = torch.Tensor(sample_time)
    
    # data_tn = df[df.samples == groups[n]].drop(columns=samples_key).values    
    data_tn = df[df.samples == times[n]].drop(columns=samples_key).values    
    data_tn = torch.Tensor(data_tn).float()
    
    if use_cuda:
        data_tn = data_tn.cuda()
        sample_time = sample_time.cuda()
            
    if autoencoder is not None and recon:
        data_tn = autoencoder.encoder(data_tn)

    generated = model(data_tn, sample_time, return_whole_sequence=True)
    if autoencoder is not None and recon:
        generated = autoencoder.decoder(generated)
    generated = to_np(generated)
    return generated


def get_cell_indexes(
    df, genes, trajectories, principal_components,
    top_n=10, where='end', start=0, palette = 'viridis', 
    samples_key='samples',  samples=None,
    cell_type_key=None, cell_types=None, use_cell_types=True
):
    '''
    Notes:
    -----------
        - `samples` refers to the timepoint sample e.g. `samples == 1` should be Boolean array
            corresponding to which rows in `df` that are of `t_1`.
            
        - `use_cell_types` determines the output shape of `top_idxs`. 
        
            + `use_cell_types=True`: `top_idxs` is a nested dictionary with structure
                ```
                    {
                        cell_type_0: {
                            gene_0: [id_0, id_1, ... id_n]
                        },
                        ...
                        cell_type_m: {...},
                    }
                ```
                Where each id is a cell of the outer cell type with the highest expression of
                the specified gene either at `t_0` (`where="start"`) or `t_n` (`where="end"`)
                e.g. cell_type_0[gene_0][0] is the id of the top cell of cell type `cell_type_0`
                expressing gene `gene_0`.
                
            + `use_cell_types=False`: `top_idxs` is a dictionary with structure
                ```
                    {                       
                        gene_0: [id_0, id_1, ... id_n],
                        gene_1: [id_0, id_1, ... id_n],
                        ...
                        gene_m: [id_0, id_1, ... id_n],                        
                    }
                ```
                Where each id is a cell (of any cell type) that has the highest expression of the
                specified gene either at `t_0` (`where="start"`) or `t_n` (`where="end"`).
    
    Arguments:
    -----------
        model (nn.Module): Trained MIOFlow model.

        df (pd.DataFrame): DataFrame of shape (n_cells, n_genes), where the ordering of 
            the columns `n_genes` corresponds to the columns of `principle_components`.
            It is assumed that the index of `df` are the cell types (but this need not be the case. 
            See `cell_types`). If there are additional columns (e.g. `samples_key`, `cell_type_key`)
            should be after the gene columns.
            
        genes (np.ndarray | list): Genes of interest to determine which cell indexes to find.
        
        trajectories (np.ndarray): Trajectories with shape (time, cells, dimensions)
        
        principal_components (np.ndarray): The principle components with shape (dimensions, n_genes).
            If used phate, can be obtained from `phate_operator.graph.data_pca.components_`
        
        top_n (int): Defaults to `10`. The number of cells to use per condition. If 
            `use_cell_types = False` this (conditions) will be the number of genes (`len(genes)`)
            otherwise it will be the number of cell types.
        
        where (str): Choices are `"start"`, and `"end"`. Defaults to `"end"`. Whether or not
            the trajectories start at `t_0` (`"start"`) or `t_n` (`"end"`). 

        start (int): Defaults to `0`. Where in `generate_tjnet_trajectories` the trajectories started.
            This is used if attempting to generate outside of `t0`. Note this works relative to `where`.
            E.g. if `where="end"` and `start=0` then this is the same as `groups[-1]`.
        
        palette (str): A Matplotlib colormap. Defaults to `"viridis"`. 
        
        samples_key (str): The name of the column in the `df` that corresponds to the time
            samples. Defaults to `"samples"`. If `df[samples_key]` throws a `KeyError` 
            either because the `df` doesnt have this column in it or typo, will resort to
            `samples` to determine this.
                        
        samples (np.ndarray | list): List of timepoints where each value corresponds to the 
            timepoint of the same row in `df`. Defaults to `None`.
        
        cell_type_key (str): The column name in the provided DataFrame `df` the corresponds to the 
            cell's cell types. Defaults to `None` which assumes the cell type is the index of the 
            `df i.e. `df.index`
        
        cell_types (np.ndarray | list): List of cell types to use from the provided DataFrame `df`.
            Defaults to `None`. If `use_cell_types = True` will attempt to figure this out from
            `cell_type_key`.
        
        use_cell_types (bool): Whether or not to use cell types.
    
    Returns:
    -----------
        genes (np.ndarray): List of genes similar to those the user passed into this function except
            in order of the columns of the provided `df`. Any genes not found in the `df` put passed in
            by the user will be removed.
            
        top_idxs (dict | dict[dict]): See notes. Dictionary or nested dictionary where leaf values are
            indicies of cells corresponding to those expressing the highest amount of specified genes.
            
        inverse (np.nddary): Reconstructed gene space from `trajectories` and `principal_components`.
            It has the shape (n_time * n_bins, n_cells, n_genes). See `generate_tjnet_trajectories`.
        
        colors (dict): Dictionary of either `{gene: color}` or `{cell_type: color}` depending on `use_cell_types`.
    '''
    # Test for valid start location
    _valid_where = 'start end'.split()
    if where not in _valid_where:
        raise ValueError(f'{where} not known. Should be one of {_valid_where}')
        
    groups = get_groups_from_df(df, samples_key, samples)

    # times = groups
    # if where == 'end':
    #     times = times[::-1]
    # times = times[start:]

    times = get_times_from_groups(groups, where, start)
    
    # Extract all of the cells at the specified either the start / end
    n = -1 if where == 'end' else 0
    # counts_n = get_sample_n_from_df(df, n, samples_key, samples, groups, drop_index=False)
    counts_n = get_sample_n_from_df(df, 0, samples_key, samples, groups=times, drop_index=False)
              
    # Filter for only known genes
    genes_mask = df.columns.isin(genes)
    # Get genes in order
    genes = df.columns[genes_mask]
        
    # Reconstruct full gene space (of just the genes we care about) 
    # from trajectories and principal components
    inverse =  np.dot(trajectories, principal_components[:, genes_mask])
                        
    if use_cell_types:
        # Try to correct for missing cell types if they are required
        cell_types = get_cell_types_from_df(df, cell_type_key, cell_types)
                
        # Get name of cell_type_key column              
        index = counts_n.columns[0] if cell_type_key is None else cell_type_key
            
        # Create colors for each cell type
        cmap = sns.color_palette(palette, n_colors=len(cell_types))
        colors = dict(zip(*[
            cell_types,
            [cmap[i] for i in range(len(cell_types))]
        ]))
        
        # For each cell type and each gene get top_n cells of that cell type expressing that gene
        top_idxs = {}
        for cell_type in cell_types:
            cells = counts_n[counts_n[index] == cell_type] 
            top_idxs[cell_type] = {}
            for gene in genes:
                top_idx = cells[gene].values.flatten().argsort()[-(top_n):]
                top_idxs[cell_type][gene] = top_idx
        
        
    else:
        # Create colors for each gene
        cmap = sns.color_palette(palette, n_colors=len(genes))
        colors = dict(zip(*[
            genes,
            [cmap[i] for i in range(len(genes))]
        ]))    
            
        # For each gene, get top_n cells expressing that gene    
        top_idxs = {}
        for gene in genes:
            top_idx = counts_n[gene].values.flatten().argsort()[-(top_n):]
            top_idxs[gene] = top_idx
            
        
    return genes, top_idxs, inverse, colors



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_models.ipynb.

# %% auto 0
__all__ = ['ToyODE', 'GrowthRateModel', 'make_model', 'Autoencoder', 'ToyModel', 'ToySDEModel', 'ConditionalODE',
           'ConditionalModel', 'ConditionalSDEModel']

# %% ../nbs/03_models.ipynb 3
import itertools
from torch.nn  import functional as F 
import torch.nn as nn
import torch
class ToyODE(nn.Module):
    """ 
    ODE derivative network
    
    feature_dims (int) default '5': dimension of the inputs, either in ambient space or embedded space.
    layer (list of int) defaulf ''[64]'': the hidden layers of the network.
    activation (torch.nn) default '"ReLU"': activation function applied in between layers.
    scales (NoneType|list of float) default 'None': the initial scale for the noise in the trajectories. One scale per bin, add more if using an adaptative ODE solver.
    n_aug (int) default '1': number of added dimensions to the input of the network. Total dimensions are features_dim + 1 (time) + n_aug. 
    
    Method
    forward (Callable)
        forward pass of the ODE derivative network.
        Parameters:
        t (torch.tensor): time of the evaluation.
        x (torch.tensor): position of the evalutation.
        Return:
        derivative at time t and position x.   
    """
    def __init__(
        self, 
        feature_dims=5,
        layers=[64],
        activation='ReLU',
        scales=None,
        n_aug=2,
        momentum_beta = 0.0
    ):
        super(ToyODE, self).__init__()
        steps = [feature_dims+1+n_aug, *layers, feature_dims]
        pairs = zip(steps, steps[1:])

        chain = list(itertools.chain(*list(zip(
            map(lambda e: nn.Linear(*e), pairs), 
            itertools.repeat(getattr(nn, activation)())
        ))))[:-1]

        self.chain = chain
        self.seq = (nn.Sequential(*chain))
        # self.initialize_final_layer()

        self.alpha = nn.Parameter(torch.tensor(scales, requires_grad=True).float()) if scales is not None else None
        self.n_aug = n_aug  

        self.momentum_beta = momentum_beta
        self.previous_v = None

    # def initialize_final_layer(self):
    #     # For the final layer of fc_g, use Xavier uniform initialization for weights
    #     # and set the bias to a small positive constant (e.g., 0.1)
    #     final_layer = self.seq[-1]
    #     if isinstance(final_layer, nn.Linear):
    #         nn.init.xavier_uniform_(final_layer.weight)
    #         nn.init.constant_(final_layer.bias, 0.1)    
        
    def forward(self, t, x): #NOTE the forward pass when we use torchdiffeq must be forward(self,t,x)
        zero = torch.tensor([0]).cuda() if x.is_cuda else torch.tensor([0])
        zeros = zero.repeat(x.size()[0],self.n_aug)
        time = t.repeat(x.size()[0],1)
        aug = torch.cat((x,time,zeros),dim=1)
        dxdt = self.seq(aug)
        if self.alpha is not None:
            z = torch.randn(x.size(),requires_grad=False).cuda() if x.is_cuda else torch.randn(x.size(),requires_grad=False)
            dxdt = dxdt + z*self.alpha[int(t-1)]

        if self.momentum_beta > 0.0:
            if self.previous_v is None or self.previous_v.shape[0] != x.shape[0]:
                self.previous_v = torch.zeros_like(dxdt)

            dxdt = self.momentum_beta * self.previous_v + (1 - self.momentum_beta) * dxdt
            self.previous_v = dxdt.detach()
        return dxdt

# %% ../nbs/03_models.ipynb 5
import itertools
from torch.nn  import functional as F 
import torch.nn as nn
import torch
import os, math, numpy as np

class GrowthRateModel(nn.Module):
    def __init__(self, feature_dims, condition_dims, layers=[64], activation='ReLU', use_time=True):
        super(GrowthRateModel, self).__init__()
        self.feature_dims = feature_dims
        self.condition_dims = condition_dims
        self.layers = layers
        self.activation = activation
        self.use_time = use_time
        self.mlp = nn.Sequential()
        self.mlp.append(nn.Linear(feature_dims + condition_dims + (1 if use_time else 0), layers[0]))
        self.mlp.append(getattr(nn, activation)())
        for i in range(len(layers)-1):
            self.mlp.append(nn.Linear(layers[i], layers[i+1]))
            self.mlp.append(getattr(nn, activation)())
        self.mlp.append(nn.Linear(layers[-1], 1)) # output the growth rate

    def forward(self, x, c=None, t=None):
        if self.condition_dims > 0:
            assert c is not None, "Condition must be provided"
            x = torch.cat([x, c], dim=1)
        if self.use_time:
            t = t.unsqueeze(1)
            x = torch.cat([x, t], dim=1)
        return self.mlp(x)

# %% ../nbs/03_models.ipynb 7
"""I want to deprecate this function."""
def make_model(
    feature_dims=5,
    layers=[64],
    output_dims=5,
    activation='ReLU',
    which='ode',
    method='rk4',
    rtol=None,
    atol=None,
    scales=None,
    n_aug=2,
    noise_type='diagonal', sde_type='ito',
    adjoint_method=None,
    use_norm=False,
    use_cuda=False,
    in_features=2, out_features=2, gunc=None,
    n_conditions=0,
    momentum_beta = 0.0,
):
    """
    Creates the 'ode' model or 'sde' model or the Geodesic Autoencoder. 
    See the parameters of the respective classes.
    """
    if which == 'ode' and n_conditions == 0:
        ode = ToyODE(feature_dims, layers, activation,scales,n_aug)
        model = ToyModel(ode,method,rtol, atol, use_norm=use_norm)
    elif which == 'ode' and n_conditions > 0:
        ode = ConditionalODE(feature_dims, n_conditions, layers, activation, scales, n_aug, momentum_beta=momentum_beta)
        model = ConditionalModel(ode, method, rtol, atol, use_norm=use_norm)
    elif which == 'sde' and n_conditions == 0:
        ode = ToyODE(feature_dims, layers, activation,scales,n_aug, momentum_beta=momentum_beta)
        gunc = ToyODE(feature_dims, layers, activation,scales,n_aug)
        model = ToySDEModel(
            ode, method, noise_type, sde_type,
            in_features=in_features, out_features=out_features, 
            gunc=gunc, 
            adjoint_method=adjoint_method, 
            use_norm=use_norm
        )
    elif which == 'sde' and n_conditions > 0:
        ode = ConditionalODE(feature_dims, n_conditions, layers, activation, scales, n_aug, momentum_beta=momentum_beta)
        gunc = ToyODE(feature_dims, layers, activation, scales, n_aug)
        model = ConditionalSDEModel(
            ode, method, noise_type, sde_type,
            in_features=in_features, out_features=out_features, 
            gunc=gunc, 
            adjoint_method=adjoint_method, 
            use_norm=use_norm
        )
    else:
        model = ToyGeo(feature_dims, layers, output_dims, activation)
    if use_cuda:
        model.cuda()
    return model 

# %% ../nbs/03_models.ipynb 8
import itertools
import torch.nn as nn
from torch.nn  import functional as F 

class Autoencoder(nn.Module):
    """ 
    Geodesic Autoencoder
    
    encoder_layers (list of int) default '[100, 100, 20]': encoder_layers[0] is the feature dimension, and encoder_layers[-1] the embedded dimension.
    decoder_layers (list of int) defaulf '[20, 100, 100]': decoder_layers[0] is the embbeded dim and decoder_layers[-1] the feature dim.
    activation (torch.nn) default '"Tanh"': activation function applied in between layers.
    use_cuda (bool) default to False: Whether to use GPU or CPU.
    
    Method
    encode
        forward pass of the encoder
        x (torch.tensor): observations
        Return:
        the encoded observations
    decode
        forward pass of the decoder
        z (torch.tensor): embedded observations
        Return:
        the decoded observations
    forward (Callable):
        full forward pass, encoder and decoder
        x (torch.tensor): observations
        Return:
        denoised observations
    """

    def __init__(
        self,
        encoder_layers = [100, 100, 20],
        decoder_layers = [20, 100, 100],
        activation = 'Tanh',
        use_cuda = False
    ):        
        super(Autoencoder, self).__init__()
        if decoder_layers is None:
            decoder_layers = [*encoder_layers[::-1]]
        device = 'cuda' if use_cuda else 'cpu'
        
        encoder_shapes = list(zip(encoder_layers, encoder_layers[1:]))
        decoder_shapes = list(zip(decoder_layers, decoder_layers[1:]))
        
        encoder_linear = list(map(lambda a: nn.Linear(*a), encoder_shapes))
        decoder_linear = list(map(lambda a: nn.Linear(*a), decoder_shapes))
        
        encoder_riffle = list(itertools.chain(*zip(encoder_linear, itertools.repeat(getattr(nn, activation)()))))[:-1]
        encoder = nn.Sequential(*encoder_riffle).to(device)
        
        decoder_riffle = list(itertools.chain(*zip(decoder_linear, itertools.repeat(getattr(nn, activation)()))))[:-1]

        decoder = nn.Sequential(*decoder_riffle).to(device)
        self.encoder = encoder
        self.decoder = decoder

        
    
    def encode(self, x):
        return self.encoder(x)

    def decode(self, z):
        return self.decoder(z)

    def forward(self, x):
        z = self.encoder(x)
        return self.decoder(z)

# %% ../nbs/03_models.ipynb 9
from torchdiffeq import odeint_adjoint as odeint
import os, math, numpy as np
import torch
import torch.nn as nn
class ToyModel(nn.Module):
    """ 
    Neural ODE
        func (nn.Module): The network modeling the derivative.
        method (str) defaulf '"rk4"': any methods from torchdiffeq.
        rtol (NoneType | float): the relative tolerance of the ODE solver.
        atol (NoneType | float): the absolute tolerance. of the ODE solver.
        use_norm (bool): if True keeps the norm of func.
        norm (list of torch.tensor): the norm of the derivative.
        
        Method
        forward (Callable)
            x (torch.tensor): the initial sample
            t (torch.tensor) time points where we suppose x is from t[0]
            return the last sample or the whole seq.      
    """
    
    def __init__(self, func, method='rk4', rtol=None, atol=None, use_norm=False):
        super(ToyModel, self).__init__()        
        self.func = func
        self.method = method
        self.rtol=rtol
        self.atol=atol
        self.use_norm = use_norm
        self.norm=[]

    def forward(self, x, t, return_whole_sequence=False):

        if self.use_norm:
            for time in t: 
                self.norm.append(torch.linalg.norm(self.func(time,x)).pow(2))
        if self.atol is None and self.rtol is None:
            x = odeint(self.func,x ,t, method=self.method)
        elif self.atol is not None and self.rtol is None:
            x = odeint(self.func,x ,t, method=self.method, atol=self.atol)
        elif self.atol is None and self.rtol is not None:
            x = odeint(self.func,x ,t, method=self.method, rtol=self.rtol)
        else: 
            x = odeint(self.func,x ,t, method=self.method, atol=self.atol, rtol=self.rtol)          
       
        x = x[-1] if not return_whole_sequence else x
        return x

    def train(self, mode=True):
        # reset the norm
        super().train(mode)
        self.norm = []

    def eval(self):
        super().eval()
        self.norm = []


# %% ../nbs/03_models.ipynb 10
from torchdiffeq import odeint_adjoint as odeint
import os, math, numpy as np
import torch
import torch.nn as nn
import torchsde

class ToySDEModel(nn.Module):
    """ 
    Neural SDE model
        func (nn.Module): drift term.
        genc (nn.Module): diffusion term.
        method (str): method of the SDE solver.
        
        Method
        forward (Callable)
            x (torch.tensor): the initial sample
            t (torch.tensor) time points where we suppose x is from t[0]
            return the last sample or the whole seq.  
    """
    
    def __init__(self, func, method='euler', noise_type='diagonal', sde_type='ito', 
    in_features=2, out_features=2, gunc=None, dt=0.1, adjoint_method=None, use_norm=False):
        super(ToySDEModel, self).__init__()        
        self.func = func
        self.method = method
        self.noise_type = noise_type
        self.sde_type = sde_type
        self.adjoint_method = adjoint_method
        self.use_norm = use_norm
        self.norm = []
        if gunc is None:
            self._gunc_args = 'y'
            self.gunc = nn.Linear(in_features, out_features)
        else:
            self._gunc_args = 't,y'
            self.gunc = gunc

        self.dt = dt
        
    def f(self, t, y):
        return self.func(t, y)

    def g(self, t, y):
        return self.gunc(t, y) if self._gunc_args == 't,y' else self.gunc(y)

    def forward(self, x, t, return_whole_sequence=False, dt=None):
        if self.use_norm:
            for time in t: 
                f_norm = torch.linalg.norm(self.func(time,x)).pow(2)
                g_norm = torch.linalg.norm(self.gunc(time,x)).pow(2)
                self.norm.append(f_norm + g_norm)
            
        dt = self.dt if self.dt is not None else 0.1 if dt is None else dt        
        if self.adjoint_method is not None:
            x = torchsde.sdeint_adjoint(self, x, t, method=self.method, adjoint_method=self.adjoint_method, dt=dt)
        else:
            x = torchsde.sdeint(self, x, t, method=self.method, dt=dt)
       
        x = x[-1] if not return_whole_sequence else x
        return x

    def train(self, mode=True):
        # reset the norm
        super().train(mode)
        self.norm = []

    def eval(self):
        super().eval()
        self.norm = []

# %% ../nbs/03_models.ipynb 11
import itertools
from torch.nn  import functional as F 
import torch.nn as nn
import torch
from torchdiffeq import odeint_adjoint as odeint
import os, math, numpy as np

"""
WARNING: Only supports local training!
"""
class ConditionalODE(nn.Module):
    """ 
    ODE derivative network
    
    feature_dims (int) default '5': dimension of the inputs, either in ambient space or embedded space.
    layer (list of int) defaulf ''[64]'': the hidden layers of the network.
    activation (torch.nn) default '"ReLU"': activation function applied in between layers.
    scales (NoneType|list of float) default 'None': the initial scale for the noise in the trajectories. One scale per bin, add more if using an adaptative ODE solver.
    n_aug (int) default '1': number of added dimensions to the input of the network. Total dimensions are features_dim + 1 (time) + n_aug. 
    
    Method
    forward (Callable)
        forward pass of the ODE derivative network.
        Parameters:
        t (torch.tensor): time of the evaluation.
        x (torch.tensor): position of the evalutation.
        Return:
        derivative at time t and position x.   
    """
    def __init__(
        self, 
        feature_dims=5,
        condition_dims=2,
        layers=[64],
        activation='ReLU',
        scales=None,
        n_aug=2,
        time_homogeneous=True,
        momentum_beta = 0.0,
    ):
        super(ConditionalODE, self).__init__()
        self.time_homogeneous = time_homogeneous
        self.time_dim = 1 if 0 else 1
        self.condition_dim = condition_dims # TODO: make it a dict instead of assuming first!
        steps = [feature_dims+self.condition_dim+self.time_dim+n_aug, *layers, feature_dims]
        pairs = zip(steps, steps[1:])

        chain = list(itertools.chain(*list(zip(
            map(lambda e: nn.Linear(*e), pairs), 
            itertools.repeat(getattr(nn, activation)())
        ))))[:-1]

        self.chain = chain
        self.seq = (nn.Sequential(*chain))
        
        self.alpha = nn.Parameter(torch.tensor(scales, requires_grad=True).float()) if scales is not None else None
        self.n_aug = n_aug
        self.momentum_beta = momentum_beta
        self.previous_v = None
        
    def set_condition(self, condition):
        self.condition = condition

    def forward(self, t, x): #NOTE the forward pass when we use torchdiffeq must be forward(self,t,x)
        zero = torch.tensor([0]).cuda() if x.is_cuda else torch.tensor([0])
        zeros = zero.repeat(x.size()[0],self.n_aug)
        time = t.repeat(x.size()[0],self.time_dim)
        aug = torch.cat((x,self.condition,time,zeros),dim=1)
        
        dxdt = self.seq(aug)
        if self.alpha is not None:
            z = torch.randn(x.size(),requires_grad=False).cuda() if x.is_cuda else torch.randn(x.size(),requires_grad=False)
            dxdt = dxdt + z*self.alpha[int(t-1)]

        if self.momentum_beta > 0.0:
            if self.previous_v is None or self.previous_v.shape[0] != x.shape[0]:
                self.previous_v = torch.zeros_like(dxdt)

            dxdt = self.momentum_beta * self.previous_v + (1 - self.momentum_beta) * dxdt
            self.previous_v = dxdt.detach()  # Update for next step 
        return dxdt
    


class ConditionalModel(nn.Module):
    """ 
    Neural ODE
        func (nn.Module): The network modeling the derivative.
        method (str) defaulf '"rk4"': any methods from torchdiffeq.
        rtol (NoneType | float): the relative tolerance of the ODE solver.
        atol (NoneType | float): the absolute tolerance. of the ODE solver.
        use_norm (bool): if True keeps the norm of func.
        norm (list of torch.tensor): the norm of the derivative.
        
        Method
        forward (Callable)
            x (torch.tensor): the initial sample
            t (torch.tensor) time points where we suppose x is from t[0]
            return the last sample or the whole seq.      
    """
    
    def __init__(self, func, method='rk4', rtol=None, atol=None, use_norm=False):
        super(ConditionalModel, self).__init__()        
        self.func = func
        self.method = method
        self.rtol=rtol
        self.atol=atol
        self.use_norm = use_norm
        self.norm=[]

    def forward(self, x, t, return_whole_sequence=False):
        # assume the last `condition_dim` dimensions of x is the condition
        x = x[:,:-self.func.condition_dim]
        c = x[:,-self.func.condition_dim:]
        self.func.set_condition(c)
        if self.use_norm:
            for time in t: 
                self.norm.append(torch.linalg.norm(self.func(time,x)).pow(2))
        if self.atol is None and self.rtol is None:
            x = odeint(self.func,x ,t, method=self.method)
        elif self.atol is not None and self.rtol is None:
            x = odeint(self.func,x ,t, method=self.method, atol=self.atol)
        elif self.atol is None and self.rtol is not None:
            x = odeint(self.func,x ,t, method=self.method, rtol=self.rtol)
        else: 
            x = odeint(self.func,x ,t, method=self.method, atol=self.atol, rtol=self.rtol)          
       
        x = x[-1] if not return_whole_sequence else x
        return x
    
    def train(self, mode=True):
        # reset the norm
        super().train(mode)
        self.norm = []

    def eval(self):
        super().eval()
        self.norm = []


class ConditionalSDEModel(ToySDEModel):
    def __init__(self, func, method='euler', noise_type='diagonal', sde_type='ito', 
    in_features=2, out_features=2, gunc=None, dt=0.1, adjoint_method=None, use_norm=False):
        super(ConditionalSDEModel, self).__init__(func, method, noise_type, sde_type, in_features, out_features, gunc, dt, adjoint_method, use_norm)

    def forward(self, x, t, return_whole_sequence=False, dt=None):
        # assume the last `condition_dim` dimensions of x is the condition
        x = x[:,:-self.func.condition_dim] # TODO: make it a dict instead of assuming first!
        c = x[:,-self.func.condition_dim:]
        self.func.set_condition(c)
        return super().forward(x, t, return_whole_sequence, dt)



# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"ode_solve": "00_ode.ipynb",
         "ODEF": "00_ode.ipynb",
         "ODEAdjoint": "00_ode.ipynb",
         "NeuralODE": "00_ode.ipynb",
         "MMD_loss": "01_losses.ipynb",
         "OT_loss": "01_losses.ipynb",
         "Density_loss": "01_losses.ipynb",
         "Local_density_loss": "01_losses.ipynb",
         "group_extract": "02_utils.ipynb",
         "sample": "02_utils.ipynb",
         "to_np": "02_utils.ipynb",
         "generate_steps": "02_utils.ipynb",
         "set_seeds": "02_utils.ipynb",
         "config_hold_out": "02_utils.ipynb",
         "config_criterion": "02_utils.ipynb",
         "get_groups_from_df": "02_utils.ipynb",
         "get_cell_types_from_df": "02_utils.ipynb",
         "get_sample_n_from_df": "02_utils.ipynb",
         "get_times_from_groups": "02_utils.ipynb",
         "ToyODE": "03_models.ipynb",
         "make_model": "03_models.ipynb",
         "Autoencoder": "03_models.ipynb",
         "ToyModel": "03_models.ipynb",
         "ToySDEModel": "03_models.ipynb",
         "plot_losses": "04_plots.ipynb",
         "plot_comparision": "04_plots.ipynb",
         "new_plot_comparisions": "04_plots.ipynb",
         "plot_gene_trends": "04_plots.ipynb",
         "train": "05_train.ipynb",
         "train_ae": "05_train.ipynb",
         "training_regimen": "05_train.ipynb",
         "__file": "06_constants.ipynb",
         "ROOT_DIR": "06_constants.ipynb",
         "DATA_DIR": "06_constants.ipynb",
         "NTBK_DIR": "06_constants.ipynb",
         "IMGS_DIR": "06_constants.ipynb",
         "RES_DIR": "06_constants.ipynb",
         "WORM_FILE": "06_constants.ipynb",
         "EB_BODIES_FILE": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_4": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_6": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_25": "06_constants.ipynb",
         "EB_BODIES_PSEUDO_82": "06_constants.ipynb",
         "DYNGEN_INFO_FILE": "06_constants.ipynb",
         "DYNGEN_EXPR_FILE": "06_constants.ipynb",
         "construct_diamond": "07_datasets.ipynb",
         "make_diamonds": "07_datasets.ipynb",
         "make_swiss_roll": "07_datasets.ipynb",
         "make_tree": "07_datasets.ipynb",
         "make_worm_data": "07_datasets.ipynb",
         "make_eb_data": "07_datasets.ipynb",
         "make_dyngen_data": "07_datasets.ipynb",
         "relabel_data": "07_datasets.ipynb",
         "rings": "07_datasets.ipynb",
         "make_rings": "07_datasets.ipynb",
         "make_jacks": "07_datasets.ipynb",
         "config_exp_logger": "08_exp.ipynb",
         "exp_log_filename": "08_exp.ipynb",
         "exp_param_filename": "08_exp.ipynb",
         "list_exps": "08_exp.ipynb",
         "gen_exp_name": "08_exp.ipynb",
         "load_exp_params": "08_exp.ipynb",
         "save_exp_params": "08_exp.ipynb",
         "setup_exp": "08_exp.ipynb",
         "is_config_subset": "08_exp.ipynb",
         "find_exps": "08_exp.ipynb",
         "test_exp": "08_exp.ipynb",
         "DiffusionDistance": "09_geo.ipynb",
         "DiffusionAffinity": "09_geo.ipynb",
         "DiffusionMap": "09_geo.ipynb",
         "PhateDistance": "09_geo.ipynb",
         "old_DiffusionDistance": "09_geo.ipynb",
         "setup_distance": "09_geo.ipynb",
         "generate_points": "10_eval.ipynb",
         "generate_trajectories": "10_eval.ipynb",
         "generate_plot_data": "10_eval.ipynb",
         "get_points_from_trajectories": "10_eval.ipynb",
         "calculate_nn": "10_eval.ipynb",
         "generate_tjnet_trajectories": "10_eval.ipynb",
         "get_cell_indexes": "10_eval.ipynb"}

modules = ["ode.py",
           "losses.py",
           "utils.py",
           "models.py",
           "plots.py",
           "train.py",
           "constants.py",
           "datasets.py",
           "exp.py",
           "geo.py",
           "eval.py"]

doc_url = "https://KrishnaswamyLab.github.io/MIOFlow/"

git_url = "https://github.com/KrishnaswamyLab/MIOFlow/tree/main/"

def custom_doc_links(name): return None
